import logging
import os.path
import subprocess
import sys
import time
from pathlib import Path
from shutil import copytree, ignore_patterns
from typing import Union, Optional

from numcodecs import Blosc

from aind_data_transfer.config_loader.imaging_configuration_loader import (
    ImagingJobConfigurationLoader,
)
from aind_data_schema.data_description import Modality
from aind_data_transfer.readers.imaging_readers import ImagingReaders
from aind_data_transfer.util.file_utils import is_cloud_url, parse_cloud_url
from aind_data_transfer.transformations.metadata_creation import (
    SubjectMetadata,
    ProceduresMetadata,
    RawDataDescriptionMetadata,
)

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


def _find_scripts_dir():
    scripts_dir = Path(os.path.abspath(__file__)).parents[3] / "scripts"
    if not scripts_dir.is_dir():
        raise Exception(f"scripts directory not found: {scripts_dir}")
    return scripts_dir


_SCRIPTS_DIR = _find_scripts_dir()

_S3_SCRIPT = _SCRIPTS_DIR / "s3_upload.py"
if not _S3_SCRIPT.is_file():
    raise Exception(f"script not found: {_S3_SCRIPT}")

_GCS_SCRIPT = _SCRIPTS_DIR / "gcs_upload.py"
if not _GCS_SCRIPT.is_file():
    raise Exception(f"script not found: {_GCS_SCRIPT}")

_OME_ZARR_SCRIPT = _SCRIPTS_DIR / "write_ome_zarr.py"
if not _OME_ZARR_SCRIPT.is_file():
    raise Exception(f"script not found: {_OME_ZARR_SCRIPT}")

_NG_LINK_SCRIPT = _SCRIPTS_DIR / "create_neuroglancer_links.py"

_SUBMIT_SCRIPT = _SCRIPTS_DIR / "cluster" / "submit.py"
if not _SUBMIT_SCRIPT.is_file():
    raise Exception(f"script not found: {_SUBMIT_SCRIPT}")


def _build_s3_cmd(
    data_src_dir: str,
    bucket: str,
    prefix: str,
    raw_image_dir_name: str,
    n_threads: int = 4,
) -> str:
    cmd = (
        f"python {_S3_SCRIPT} "
        f"--input={data_src_dir} "
        f"--bucket={bucket} "
        f"--s3_path={prefix} "
        f"--nthreads={n_threads} "
        f"--recursive "
        f"--exclude_dirs={raw_image_dir_name}"
    )
    return cmd


def _build_gcs_cmd(
    data_src_dir: str,
    bucket: str,
    prefix: str,
    raw_image_dir_name: str,
    n_threads: int = 4,
) -> str:
    cmd = (
        f"python {_GCS_SCRIPT} "
        f"--input={data_src_dir} "
        f"--bucket={bucket} "
        f"--gcs_path={prefix} "
        f"--nthreads={n_threads} "
        f"--recursive "
        f"--method=python "
        f"--exclude_dirs={raw_image_dir_name} "
    )
    return cmd


def _build_ome_zar_cmd(
    raw_image_dir: str, zarr_out: str, job_configs: dict, bkg_im_dir: Optional[Union[str, Path]] = None
) -> str:
    compression_opts = _resolve_compression_options(job_configs)
    job_opts = job_configs["transcode_job"]
    job_cmd = (
        f"python {_OME_ZARR_SCRIPT} "
        f"--input={raw_image_dir} "
        f"--output={zarr_out} "
        f"--codec={compression_opts['cname']} "
        f"--clevel={compression_opts['clevel']} "
        f"--n_levels={job_opts['n_levels']} "
        f"--chunk_size={job_opts['chunk_size']} "
        f"--scale_factor=2 "
        f"--deployment=slurm"
    )
    if "chunk_shape" in job_opts and job_opts["chunk_shape"]:
        chunks = " ".join(str(el) for el in job_opts["chunk_shape"])
        job_cmd += f" --chunk_shape {chunks}"
    if "exclude" in job_opts and job_opts["exclude"]:
        exclusions = " ".join(job_opts["exclude"])
        job_cmd += f" --exclude {exclusions}"
    if "resume" in job_opts and job_opts["resume"]:
        job_cmd += " --resume"
    if "voxsize" in job_opts and job_opts["voxsize"] != "":
        voxsize = job_opts["voxsize"]
        job_cmd += f" --voxsize {voxsize}"
    if bkg_im_dir is not None:
        LOGGER.info("Doing background subtraction")
        job_cmd += f" --bkg_img_dir {str(bkg_im_dir)}"
    return job_cmd


def _build_submit_cmd(
    job_cmd: str, job_configs: dict, wait: bool = False
) -> str:
    submit_args = job_configs["transcode_job"]["submit_args"]
    # FIXME: necessary to wrap job_cmd in quotes
    submit_cmd = f'python {_SUBMIT_SCRIPT} generate-and-launch-run --job_cmd="{job_cmd}"'
    for k, v in submit_args.items():
        submit_cmd += f" --{k}={v}"
    if wait:
        submit_cmd += " --wait"
    return submit_cmd


def _resolve_compression_options(job_configs: dict) -> dict:
    opts = {}

    try:
        compressor_kwargs = job_configs["transcode_job"]["compressor"][
            "kwargs"
        ]
    except KeyError:
        compressor_kwargs = {}

    opts["cname"] = compressor_kwargs.get("cname", "zstd")
    opts["clevel"] = compressor_kwargs.get("clevel", 1)
    opts["shuffle"] = compressor_kwargs.get("shuffle", Blosc.SHUFFLE)

    return opts


def main():
    job_configs = ImagingJobConfigurationLoader().load_configs(sys.argv[1:])

    data_src_dir = Path(job_configs["endpoints"]["raw_data_dir"])
    dest_data_dir = job_configs["endpoints"]["dest_data_dir"]
    if dest_data_dir.endswith("/"):
        # remove trailing slash
        dest_data_dir = dest_data_dir[:-1]

    reader = ImagingReaders.get_reader_name(data_src_dir)
    raw_image_dir = ImagingReaders.get_raw_data_dir(reader, data_src_dir)

    LOGGER.info(f"Transferring data to {dest_data_dir}")

    raw_image_dir_name = Path(raw_image_dir).name

    wait = job_configs["jobs"]["create_ng_link"]
    if wait:
        LOGGER.info(
            "Will wait for job to terminate before continuing execution"
        )

    zarr_out = dest_data_dir + "/" + raw_image_dir_name + ".zarr"
    if job_configs["jobs"]["transcode"]:
        bkg_im_dir = None
        if job_configs["jobs"]["background_subtraction"]:
            bkg_im_dir = data_src_dir / "derivatives"
            if not bkg_im_dir.is_dir():
                raise Exception(f"background image directory not found: {bkg_im_dir}")
            LOGGER.info(f"Using background image directory: {bkg_im_dir}")
        job_cmd = _build_ome_zar_cmd(raw_image_dir, zarr_out, job_configs, bkg_im_dir)
        submit_cmd = _build_submit_cmd(job_cmd, job_configs, wait)
        subprocess.run(submit_cmd, shell=True)
        LOGGER.info("Submitted transcode job to cluster")

    if job_configs["jobs"]["create_ng_link"]:
        ng_link_cmd = (
            f"python {_NG_LINK_SCRIPT} "
            f"--input={zarr_out} "
            f"--output={data_src_dir} "
            f"--vmin={job_configs['create_ng_link_job']['vmin']} "
            f"--vmax={job_configs['create_ng_link_job']['vmax']}"
        )
        subprocess.run(ng_link_cmd, shell=True)
        output_json = data_src_dir / "process_output.json"
        if not output_json.is_file():
            LOGGER.error(
                f"Creating neuroglancer link failed; {output_json} was not created"
            )

    if job_configs["jobs"]["create_metadata"]:
        metadata_service_url = job_configs["endpoints"]["metadata_service_url"]
        subject_id = job_configs["data"]["subject_id"]
        subject_metadata = SubjectMetadata.from_service(
            domain=metadata_service_url,
            subject_id=subject_id,
        )
        subject_metadata.write_to_json(
            os.path.join(data_src_dir, subject_metadata.output_filename)
        )

        procedures_metadata = ProceduresMetadata.from_service(
            domain=metadata_service_url,
            subject_id=subject_id,
        )
        procedures_metadata.write_to_json(
            os.path.join(data_src_dir, procedures_metadata.output_filename)
        )

        data_description_metadata = RawDataDescriptionMetadata.from_inputs(
            name=Path(data_src_dir).name, modality=[Modality.SPIM]
        )
        data_description_metadata.write_to_json(
            os.path.join(
                data_src_dir, data_description_metadata.output_filename
            )
        )

    if job_configs["jobs"]["upload_aux_files"]:
        LOGGER.info("Uploading auxiliary data")
        t0 = time.time()
        if is_cloud_url(dest_data_dir):
            provider, bucket, prefix = parse_cloud_url(dest_data_dir)
            if provider == "s3://":
                cmd = _build_s3_cmd(
                    data_src_dir, bucket, prefix, raw_image_dir_name
                )
            elif provider == "gs://":
                cmd = _build_gcs_cmd(
                    data_src_dir, bucket, prefix, raw_image_dir_name
                )
            else:
                raise Exception(f"Unsupported cloud storage: {provider}")
            subprocess.run(cmd, shell=True)
        else:
            copytree(
                data_src_dir,
                dest_data_dir,
                ignore=ignore_patterns(raw_image_dir_name),
                dirs_exist_ok=True
            )
        LOGGER.info(
            f"Finished uploading auxiliary data, took {time.time() - t0}"
        )


if __name__ == "__main__":
    main()



"""Module that contains the API to retrieve a Compressor for Ephys Data.
"""
import logging
import shutil
from enum import Enum
from pathlib import Path
from typing import Optional

import spikeinterface.preprocessing as spre
from aind_data_schema.data_description import Modality
from aind_data_schema.processing import ProcessName
from numcodecs import Blosc
from numpy import memmap
from pydantic import Field
from wavpack_numcodecs import WavPack

from aind_data_transfer.config_loader.base_config import ModalityConfigs
from aind_data_transfer.readers.ephys_readers import DataReader, EphysReaders
from aind_data_transfer.util.npopto_correction import (
    correct_np_opto_electrode_locations,
)
from aind_data_transfer.writers.ephys_writers import EphysWriters


class CompressorName(Enum):
    """Enum for compression algorithms a user can select"""

    BLOSC = Blosc.codec_id
    WAVPACK = "wavpack"


class EcephysCompressionParameters(ModalityConfigs):
    """Extra configs for Ecephys upload job."""

    modality = Modality.ECEPHYS

    # Override these values from the base settings
    process_name: ProcessName = Field(
        default=ProcessName.EPHYS_PREPROCESSING,
        description="Type of processing performed on the raw data source.",
        title="Process Name",
        const=True,
    )

    data_reader: DataReader = Field(
        default=DataReader.OPENEPHYS,
        description="Type of reader to use to read the data source.",
        title="Data Reader",
    )

    # Clip settings
    clip_n_frames: int = Field(
        default=100,
        description="Number of frames to clip the data.",
        title="Clip N Frames",
    )
    # Compress settings
    compress_write_output_format: str = Field(
        default="zarr",
        description=(
            "Output format for compression. Currently, only zarr supported."
        ),
        title="Write Output Format",
        const=True,
    )
    compress_max_windows_filename_len: int = Field(
        default=150,
        description=(
            "Windows OS max filename length is 256. The zarr write will "
            "raise an error if it detects that the destination directory has "
            "a long name."
        ),
        title="Compress Max Windows Filename Len",
    )
    compressor_name: CompressorName = Field(
        default=CompressorName.WAVPACK,
        description="Type of compressor to use.",
        title="Compressor Name.",
    )
    compressor_kwargs: dict = Field(
        default={"level": 3},
        description="Arguments to be used for the compressor.",
        title="Compressor Kwargs",
    )
    compress_job_save_kwargs: dict = Field(
        default={"n_jobs": -1},  # -1 to use all available cpu cores.
        description="Arguments for recording save method.",
        title="Compress Job Save Kwargs",
    )
    compress_chunk_duration: str = Field(
        default="1s",
        description="Duration to be used for chunks.",
        title="Compress Chunk Duration",
    )

    # Scale settings
    scale_num_chunks_per_segment: int = Field(
        default=100,
        description="Num of chunks per segment to scale.",
        title="Scale Num Chunks Per Segment",
    )
    scale_chunk_size: int = Field(
        default=10000,
        description="Chunk size to scale.",
        title="Scale Chunk Size",
    )


class EphysCompressors:
    """This class contains the methods to retrieve a compressor, and to scale
    a read block by lsb and median values.
    """

    compressors = [member.value for member in CompressorName]

    def __init__(
        self,
        job_configs: EcephysCompressionParameters,
        behavior_dir: Optional[Path] = None,
        log_level: str = "WARNING",
    ):
        """Class constructor"""
        self.job_configs = job_configs
        self.behavior_dir = behavior_dir
        self._instance_logger = (
            logging.getLogger(__name__)
            .getChild(self.__class__.__name__)
            .getChild(str(id(self)))
        )
        self._instance_logger.setLevel(log_level)

    @staticmethod
    def get_compressor(compressor_name, **kwargs):
        """
        Retrieve a compressor for a given name and optional kwargs.
        Args:
            compressor_name (str): Matches one of the names Compressors enum
            **kwargs (dict): Options to pass into the Compressor
        Returns:
            An instantiated compressor class.
        """
        if compressor_name == CompressorName.BLOSC.value:
            return Blosc(**kwargs)
        elif compressor_name == CompressorName.WAVPACK.value:
            return WavPack(**kwargs)
        else:
            raise Exception(
                f"Unknown compressor. Please select one of "
                f"{EphysCompressors.compressors}"
            )

    @staticmethod
    def scale_read_blocks(
        read_blocks,
        num_chunks_per_segment=100,
        chunk_size=10000,
    ):
        """
        Scales a read_block. A read_block is dict of
        {'recording', 'block_index', 'stream_name'}.
        Args:
            read_blocks (iterable): A generator of read_blocks
            num_chunks_per_segment (int):
            chunk_size (int):
        Returns:
            A generated scaled_read_block. A dict of
            {'scaled_recording', 'block_index', 'stream_name'}.

        """
        for read_block in read_blocks:
            # We don't need to scale the NI-DAQ recordings
            # TODO: Convert this to regex matching?
            if (
                EphysReaders.RecordingBlockPrefixes.nidaq.value
                in read_block["stream_name"]
            ):
                rec_to_compress = read_block["recording"]
            else:
                rec_to_compress = spre.correct_lsb(
                    read_block["recording"],
                    num_chunks_per_segment=num_chunks_per_segment,
                    chunk_size=chunk_size,
                )
            yield (
                {
                    "scaled_recording": rec_to_compress,
                    "experiment_name": read_block["experiment_name"],
                    "stream_name": read_block["stream_name"],
                }
            )

    def _copy_and_clip_data(
        self,
        dst_dir,
        stream_gen,
    ):
        """
        Copies the raw data to a new directory with the .dat files clipped to
        just a small number of frames. This allows someone to still use the
        spikeinterface api on the clipped data set.
        Parameters
        ----------
        dst_dir : Path
          Desired location for clipped data set
        stream_gen : dict
          A dict with
            'data': memmap(dat file),
              'relative_path_name': path name of raw data
                to new dir correctly
              'n_chan': number of channels.
        Returns
        -------
        None
          Moves some directories around.

        """

        # first: copy everything except .dat files and files in behavior dir
        if self.behavior_dir is None:
            patterns_to_ignore = ["*.dat"]
        else:
            behavior_glob = self.behavior_dir / "*"
            patterns_to_ignore = ["*.dat", str(behavior_glob)]
        shutil.copytree(
            self.job_configs.source,
            dst_dir,
            ignore=shutil.ignore_patterns(*patterns_to_ignore),
        )
        # second: copy clipped dat files
        for stream in stream_gen:
            data = stream["data"]
            rel_path_name = stream["relative_path_name"]
            n_chan = stream["n_chan"]
            dst_raw_file = dst_dir / rel_path_name
            dst_data = memmap(
                dst_raw_file,
                dtype="int16",
                shape=(self.job_configs.clip_n_frames, n_chan),
                order="C",
                mode="w+",
            )
            dst_data[:] = data[: self.job_configs.clip_n_frames]

    def compress_raw_data(self, temp_dir: Path) -> None:
        """If compress data is set to False, the data will be uploaded to s3.
        Otherwise, it will be compressed to zarr, stored in temp_dir, and
        uploaded later."""

        # Correct NP-opto electrode positions:
        # correction is skipped if Neuropix-PXI version > 0.4.0
        # It'd be nice if the original data wasn't modified.
        correct_np_opto_electrode_locations(self.job_configs.source)
        # Clip the data
        self._instance_logger.info(
            "Clipping source data. This may take a minute."
        )
        if self.job_configs.number_id is None:
            clipped_data_path = temp_dir / "ecephys_clipped"
        else:
            clipped_data_path = (
                temp_dir / f"ecephys_clipped{self.job_configs.number_id}"
            )
        streams_to_clip = EphysReaders.get_streams_to_clip(
            self.job_configs.data_reader.value,
            self.job_configs.source,
        )
        self._copy_and_clip_data(
            dst_dir=clipped_data_path,
            stream_gen=streams_to_clip,
        )

        self._instance_logger.info("Finished clipping source data.")

        # Compress the data
        self._instance_logger.info("Compressing source data.")
        if self.job_configs.number_id is None:
            compressed_data_path = temp_dir / "ecephys_compressed"
        else:
            compressed_data_path = (
                temp_dir / f"ecephys_compressed{self.job_configs.number_id}"
            )
        read_blocks = EphysReaders.get_read_blocks(
            self.job_configs.data_reader.value,
            self.job_configs.source,
        )
        compressor = EphysCompressors.get_compressor(
            self.job_configs.compressor_name.value,
            **self.job_configs.compressor_kwargs,
        )
        scaled_read_blocks = EphysCompressors.scale_read_blocks(
            read_blocks=read_blocks,
            num_chunks_per_segment=(
                self.job_configs.scale_num_chunks_per_segment
            ),
            chunk_size=self.job_configs.scale_chunk_size,
        )
        EphysWriters.compress_and_write_block(
            read_blocks=scaled_read_blocks,
            compressor=compressor,
            output_dir=compressed_data_path,
            max_windows_filename_len=(
                self.job_configs.compress_max_windows_filename_len
            ),
            output_format=self.job_configs.compress_write_output_format,
            job_kwargs=self.job_configs.compress_job_save_kwargs,
        )
        self._instance_logger.info("Finished compressing source data.")

        return None



import unittest
import os

import numpy as np
import dask.array as da

from aind_data_transfer.transformations.flatfield_correction import BkgSubtraction


class TestBkgSubtraction(unittest.TestCase):

    def test_subtract(self):
        im = da.from_array(np.full((64, 64, 64), fill_value=100, dtype=int), chunks=(16, 16, 16))
        bkg_im = da.from_array(np.full((32, 32), fill_value=50, dtype=int), chunks=(16, 16))
        result = BkgSubtraction.subtract(im, bkg_im)
        self.assertIsInstance(result, da.Array)
        self.assertEqual(result.shape, im.shape)
        self.assertEqual(result.sum().compute(), np.product(im.shape) * 50)

    def test_get_bkg_path(self):
        bkg_im_dir = "path/to/background_images/"

        with self.assertRaises(ValueError):
            BkgSubtraction.get_bkg_path("invalid_tile_path", bkg_im_dir)

        # Test with a valid tile path
        raw_tile_path = "/path/to/tiles/tile_X_0000_Y_0000_Z_0000_ch_488.ims"
        expected_bkg_path = os.path.join(bkg_im_dir, "bkg_tile_X_0000_Y_0000_Z_0000.tiff")
        result = BkgSubtraction.get_bkg_path(raw_tile_path, bkg_im_dir)
        self.assertEqual(result, expected_bkg_path)


if __name__ == '__main__':
    unittest.main()



import logging
import os
import platform
import shutil

import numpy as np

from aind_data_transfer.transformations.generic_compressors import (
    VideoCompressor,
)

DEFAULT_MAX_WINDOWS_FILENAME_LEN = 150


class EphysWriters:
    """This class contains the methods to write ephys data."""

    @staticmethod
    def compress_and_write_block(
        read_blocks,
        compressor,
        output_dir,
        job_kwargs,
        output_format="zarr",
        max_windows_filename_len=DEFAULT_MAX_WINDOWS_FILENAME_LEN,
    ):
        """
        Compress and write read_blocks.
        Args:
            read_blocks (iterable dict):
              Either [{'recording', 'block_index', 'stream_name'}] or
              [{'scale_recording', 'block_index', 'stream_name'}].
            compressor (obj): A compressor class
            output_dir (Path): Output directory to write compressed data
            job_kwargs (dict): Recording save job kwargs.
            output_format (str): Defaults to zarr
            max_windows_filename_len (int): Warn if base file names are larger
              than this.

        Returns:
            Nothing. Writes data to a folder.
        """
        if job_kwargs["n_jobs"] == -1:
            job_kwargs["n_jobs"] = os.cpu_count()

        if max_windows_filename_len is None:
            max_windows_filename_len = DEFAULT_MAX_WINDOWS_FILENAME_LEN

        for read_block in read_blocks:
            if "recording" in read_block:
                rec = read_block["recording"]
            else:
                rec = read_block["scaled_recording"]
            experiment_name = read_block["experiment_name"]
            stream_name = read_block["stream_name"]
            zarr_path = output_dir / f"{experiment_name}_{stream_name}.zarr"
            if (
                platform.system() == "Windows"
                and len(str(zarr_path)) > max_windows_filename_len
            ):
                raise Exception(
                    f"File name for zarr path is too long "
                    f"({len(str(zarr_path))})"
                    f" and might lead to errors. Use a shorter destination "
                    f"path."
                )
            _ = rec.save(
                format=output_format,
                zarr_path=zarr_path,
                compressor=compressor,
                **job_kwargs,
            )

    @staticmethod
    def copy_and_clip_data(
        src_dir,
        dst_dir,
        stream_gen,
        behavior_dir=None,
        video_encryption_key=None,
        n_frames=100,
    ):
        """
        Copies the raw data to a new directory with the .dat files clipped to
        just a small number of frames. This allows someone to still use the
        spikeinterface api on the clipped data set. Also zips and encrypts
        video folder.
        Parameters
        ----------
        src_dir : Path
          Location of raw data
        dst_dir : Path
          Desired location for clipped data set
        stream_gen : dict
          A dict with
            'data': np.memmap(dat file),
              'relative_path_name': path name of raw data
                to new dir correctly
              'n_chan': number of channels.
        behavior_dir: Path
          Location of videos files to compress. If None, will check if src_dir
          contains a Video or video folder in its root.
        video_encryption_key : Optional[str]
          Password to use to encrypt video files. Default is None.
        n_frames : int
          Number of frames to clip data to. Default is 100.

        Returns
        -------
        None
          Moves some directories around.

        """
        # first: copy everything except .dat files
        shutil.copytree(
            src_dir, dst_dir, ignore=shutil.ignore_patterns("*.dat")
        )
        # second: copy clipped dat files
        for stream in stream_gen:
            data = stream["data"]
            rel_path_name = stream["relative_path_name"]
            n_chan = stream["n_chan"]
            dst_raw_file = dst_dir / rel_path_name
            dst_data = np.memmap(
                dst_raw_file,
                dtype="int16",
                shape=(n_frames, n_chan),
                order="C",
                mode="w+",
            )
            dst_data[:] = data[:n_frames]

        # third: check if videos directory exists
        new_videos_path = dst_dir / ".." / "behavior"
        if behavior_dir is not None:
            videos_path = behavior_dir
            shutil.copytree(videos_path, new_videos_path)
        elif os.path.isdir(dst_dir / "Videos"):
            videos_path = dst_dir / "Videos"
            shutil.move(videos_path, new_videos_path)
        elif os.path.isdir(dst_dir / "videos"):
            videos_path = dst_dir / "videos"
            shutil.move(videos_path, new_videos_path)
        else:
            videos_path = None

        # Log a warning if no videos path found
        if videos_path is None:
            logging.warning("No videos found!")
        else:
            # Compress and optionally encrypt
            video_compressor = VideoCompressor(
                encryption_key=video_encryption_key
            )
            video_compressor.compress_all_videos_in_dir(new_videos_path)



"""
Submit jobs to a cluster.
Adapted from from https://github.com/JaneliaSciComp/spark-janelia
"""

import argparse
import collections
import datetime
import os
import shutil
import string
import time
from pathlib import Path


class SlurmTemplate(string.Template):
    delimiter = "@"
    idpattern = r"[a-z][_a-z0-9]*"


def render_template(template_path, to_path, **kwargs):
    with open(template_path, "r") as template_file:
        template_contents = template_file.read()

    template = SlurmTemplate(template_contents)
    populated_template = template.substitute(kwargs)

    with open(to_path, "w") as file:
        print(populated_template, file=file)

    return to_path


def write_slurm_scripts(args, run_info):
    template_path = os.path.join(
        Path(os.path.dirname(__file__)).parent.parent.absolute(),
        "templates/queue_slurm_job.sh",
    )
    script_path = render_template(
        template_path,
        run_info.launch_script,
        conda_activate=args.conda_activate,
        conda_env=args.conda_env,
        job_cmd=args.job_cmd,
        mail_user=args.mail_user,
        job_log_dir=run_info.logs_dir,
        walltime=args.walltime,
        partition=args.queue,
        nodes=args.nodes,
        ntasks_per_node=args.ntasks_per_node,
        cpus_per_task=args.cpus_per_task,
        mem_per_cpu=args.mem_per_cpu,
        tmp_space=args.tmp_space,
        dask_conf_file=run_info.dask_conf_file,
    )
    os.chmod(script_path, 0o755)


def copy_dask_config(run_info):
    conf_file = (
        Path(__file__).parent.parent.parent / "conf/dask/dask-config.yaml"
    )
    assert conf_file.is_file()
    # Copy to the run directory.
    # The DASK_CONFIG environment variable is set to this path
    shutil.copyfile(conf_file, run_info.dask_conf_file)


RunInfo = collections.namedtuple(
    "RunInfo",
    "dir, logs_dir, scripts_dir, launch_script, "
    "job_name_prefix, dask_conf_file",
)


def create_run(args):
    timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
    run_dir = f"{args.run_parent_dir}/{timestamp}"

    try:
        os.makedirs(run_dir)
    except FileExistsError:
        # retry once to work around rare case of concurrent run by same user
        time.sleep(1)
        timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        run_dir = f"{args.run_parent_dir}/{timestamp}"

    run_logs_dir = f"{run_dir}/logs"
    run_scripts_dir = f"{run_dir}/scripts"
    run_conf_dir = f"{run_dir}/conf"
    launch_script = f"{run_scripts_dir}/queue-slurm-jobs.sh"

    dask_conf_file = f"{run_conf_dir}/dask-config.yaml"

    user_name = os.getenv("USER")
    job_name_prefix = f"slurm_{user_name}_{timestamp}"

    os.makedirs(run_logs_dir)
    os.makedirs(run_scripts_dir)
    os.makedirs(run_conf_dir)

    print(f"\nCreated:\n {run_logs_dir}\n  {run_scripts_dir}\n")

    return RunInfo(
        run_dir,
        run_logs_dir,
        run_scripts_dir,
        launch_script,
        job_name_prefix,
        dask_conf_file,
    )


def parse_args():
    home_dir = os.getenv("HOME")

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "task",
        choices=["generate-run", "generate-and-launch-run"],
        help="use generate-run if you want to review scripts before launching",
    )
    parser.add_argument(
        "-j",
        "--job_cmd",
        type=str,
        required=True,
        help="the job command to submit to the scheduler",
    )
    parser.add_argument(
        "--run_parent_dir",
        type=str,
        default=f"{home_dir}/.slurm",
        help="directory to store all logs, scripts, "
        "and config files for a run",
    )
    parser.add_argument(
        "--conda_activate",
        type=str,
        help="path to conda env activation script",
    )
    parser.add_argument(
        "--conda_env", type=str, help="name of conda environment to activate"
    )
    parser.add_argument(
        "--ntasks_per_node",
        type=int,
        default=1,
        help="number of tasks per node",
    )
    parser.add_argument(
        "--nodes", type=int, default=1, help="number of HPC nodes"
    )
    parser.add_argument(
        "--cpus_per_task",
        type=int,
        default=1,
        help="number of cpus per job",
    )
    parser.add_argument(
        "--mem_per_cpu", type=int, default=500, help="memory per cpu in MB"
    )
    parser.add_argument(
        "--tmp_space",
        type=str,
        default="512MB",
        help="amount of space on node-local storage for "
        "Dask worker spilling, e.g., 4GB",
    )
    parser.add_argument(
        "--walltime", type=str, default="00:10:00", help="job walltime"
    )
    parser.add_argument(
        "--queue", type=str, default="aind", help="queue for scheduler"
    )
    parser.add_argument(
        "-m",
        "--mail_user",
        type=str,
        required=True,
        help="notify user of job status",
    )
    parser.add_argument(
        "--deployment",
        type=str,
        choices=["slurm"],
        default="slurm",
        help="which cluster manager to use",
    )
    parser.add_argument(
        "--local_dir",
        type=str,
        default="/scratch/fast/$SLURM_JOB_ID",
        help="local directory for dask worker file spilling. "
        "This should be fast, node-local storage.",
    )
    parser.add_argument(
        "--wait",
        default=False,
        action="store_true",
        help="do not exit until the submitted job terminates. See the --wait flag for sbatch.",
    )

    args = parser.parse_args()

    return args


def main():
    args = parse_args()

    run_info = create_run(args)

    if args.deployment == "slurm":
        write_slurm_scripts(args, run_info)
    else:
        raise ValueError("Only Slurm is currently supported")

    copy_dask_config(run_info)

    if args.task == "generate-and-launch-run":
        cmd = f"sbatch "
        if args.wait:
            cmd += "--wait "
        cmd += run_info.launch_script
        print(f"Running:\n  {cmd}\n")
        os.system(cmd)
    else:
        print(f"To launch jobs, run:\n  {run_info.launch_script}\n")


if __name__ == "__main__":
    main()



import os
import unittest
from unittest import mock

from google.auth.exceptions import DefaultCredentialsError

from aind_data_transfer.gcs import GCSUploader


class TestGCSUploader(unittest.TestCase):
    """Tests methods defined in GCSUploader class"""

    @mock.patch.dict(os.environ, {"GOOGLE_APPLICATION_CREDENTIALS": "NONE"})
    def test_gcs_uploader_init(self):
        """Checks that the GCSUploader can be instantiated. Since GCSUploader
        initializes a cloud storage client, we'll test that it throws an
        error when the credentials don't exist.
        """

        with self.assertRaises(DefaultCredentialsError):
            GCSUploader(bucket_name=None)


if __name__ == "__main__":
    unittest.main()



"""
Script that uploads smartspim datasets. It is designed
to be automatic by using a processing_manifest.json
inside of each dataset and a YAML configuration file
located in the conf/ folder.
"""
import glob
import json
import logging
import os
import sys
import time
import warnings
from datetime import datetime
from pathlib import Path
from typing import Optional, Tuple, Union

import s3_upload
import yaml
from argschema import ArgSchema, ArgSchemaParser
from argschema.fields import Dict, InputFile, Int, List, Str
from validate_datasets import validate_dataset

from aind_data_transfer.readers.imaging_readers import SmartSPIMReader
from aind_data_transfer.util import file_utils
from aind_data_transfer.writers.imaging_writers import SmartSPIMWriter

warnings.filterwarnings("ignore")

CURR_DATE = datetime.now().strftime("%Y-%m-%d")

# Printing logs in .cronjob folder in my root folder structure
file_path = Path(
    os.path.dirname(os.path.realpath(__file__))
).parent.parent.absolute()

LOGS_FOLDER = f"{file_path}/.cronjob/logs_{CURR_DATE}"
CURR_DATE_TIME = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")

# Creates logs folder, if it does not exist
file_utils.create_folder(LOGS_FOLDER)

logging.basicConfig(
    level=logging.DEBUG,
    format="%(asctime)s - %(levelname)s : %(message)s",
    datefmt="%Y-%m-%d %H:%M",
    handlers=[
        logging.StreamHandler(),
        logging.FileHandler(
            f"{LOGS_FOLDER}/smartspim_logs_{CURR_DATE_TIME}.log", "a"
        ),
    ],
    force=True,
)
logging.disable("DEBUG")
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

PathLike = Union[str, Path]


class ConfigFile(ArgSchema):
    """YAML configuration file"""

    config_file = InputFile(
        required=True,
        metadata={"description": "Path to the YAML config file."},
    )


class CopyDatasets(ArgSchema):
    """
    Argschema parameters to copy smartspim
    datasets
    """

    transfer_type = Dict(
        required=True,
        metadata={"description": "Configuration for the transfer type"},
    )

    metadata_service_domain = Str(
        required=True,
        metadata={"description": "Metadata service domain"},
        dump_default="",
    )

    codeocean_credentials_path = Str(
        required=True,
        metadata={
            "description": "Path where the code ocean credentials are saved"
        },
        dump_default="",
    )

    co_capsule_id = Str(
        required=True,
        metadata={"description": "Capsule id of smartspim pipeline"},
    )

    root_folder = Str(
        required=True,
        metadata={
            "description": "Folder where smartspim datasets are located"
        },
    )

    dest_data_dir = Str(
        required=False,
        metadata={
            "description": "Relative path to bucket where the data will be stored"
        },
        dump_default="",
    )

    s3_bucket = Str(
        required=True,
        metadata={"description": "S3 bucket where the data will be stored"},
    )

    nthreads = Int(
        required=True,
        metadata={"description": "Number of threads to upload data"},
        dump_default=1,
    )

    exiftool_path = Str(
        required=True,
        metadata={"description": "Path to exiftool"},
    )

    info_manager_path = Str(
        required=False,
        metadata={
            "description": "Path to output status of smartspim datasets"
        },
        dump_default=None,
    )


def get_default_config(filename: str) -> dict:
    """
    Get default configuration from a YAML file.

    Parameters
    ------------------------
    filename: str
        String where the YAML file is located.

    Returns
    ------------------------
    Dict
        Dictionary with the configuration
    """

    filename = Path(os.path.dirname(__file__)).joinpath(filename)

    config = None
    try:
        with open(filename, "r") as stream:
            config = yaml.safe_load(stream)
    except Exception as error:
        raise error

    return config


def get_smartspim_folders_with_file(
    root_folder: PathLike, search_file: str
) -> Tuple[List]:
    """
    Get the smartspim folder that
    contain the search filename in
    the top level directory.

    Parameters
    ----------
    root_folder: PathLike
        Path to the root folder where
        the smartspim datasets are located

    search_file: str
        Filename to search

    Returns
    ----------
    Tuple[List[str]]
        Returns a tuple with two
        positions. The first one corresponds
        to the smartspim datasets that contain
        the provided filename, while the second
        one provides the rejected datasets
    """

    root_folder = Path(root_folder)
    raw_smartspim_datasets = SmartSPIMReader.read_raw_smartspim_folders(
        root_folder
    )

    raw_datasets_accepted = []
    raw_datasets_rejected = []

    for raw_smartspim_dataset in raw_smartspim_datasets:
        file_path = root_folder.joinpath(raw_smartspim_dataset, search_file)

        if os.path.exists(file_path):
            raw_datasets_accepted.append(raw_smartspim_dataset)

        else:
            raw_datasets_rejected.append(raw_smartspim_dataset)

    return raw_datasets_accepted, raw_datasets_rejected


def organize_datasets(
    root_folder: PathLike,
    datasets: list,
    metadata_service: str,
    config_filename: str = "processing_manifest.json",
) -> Tuple[list, list]:
    """
    This function organizes the smartspim
    datasets

    Parameters
    ------------------------
    root_folder: PathLike
        Root folder where these datasets
        are stored

    datasets: List
        List with the smartspim dataset
        paths.

    Returns
    --------------
    Tuple[List[str], List[dict]]
        Tuple that contains in the first position
        a list with the new paths for the smartspim
        datasets, and in the second position a list
        with the pipeline configuration of those
        datasets
    """
    new_dataset_paths = []
    ready_datasets = []

    if isinstance(datasets, list):
        for dataset in datasets:
            dataset_path = Path(root_folder).joinpath(dataset)

            if not os.path.isdir(dataset_path):
                logger.info(f"Dataset {dataset_path} does not exist.")
                continue

            json_path_config = dataset_path.joinpath(config_filename)
            logger.info(f"Reading {json_path_config}")

            # Reading `processing_manifest.json`
            dataset_config = file_utils.read_json_as_dict(json_path_config)

            if "dataset_status" not in dataset_config:
                logger.error(
                    f"Ignoring dataset {dataset_path}, it does not have the dataset status attribute or json does not exist."
                )

            elif dataset_config["dataset_status"].casefold() == "pending":
                logger.info(f"Validating dataset {dataset_path}")
                if validate_dataset(dataset_path):
                    dataset_config["path"] = dataset_path
                    ready_datasets.append(dataset_config)

                else:
                    logger.info(
                        f"Dataset {dataset_path} has problems in tiles!\n"
                    )
            else:
                logger.error(
                    f"Ignoring dataset {dataset_path}, it does not have the PENDING tag in dataset_status."
                )

        # Organizing smartspim folders
        smartSPIM_writer = SmartSPIMWriter(
            dataset_paths=ready_datasets,
            metadata_domain=metadata_service,
        )

        (
            new_dataset_paths,
            ignored_datasets,
        ) = smartSPIM_writer.prepare_datasets(mode="move")

    return new_dataset_paths, ready_datasets


def get_smartspim_default_config() -> dict:
    """
    Returns the default config
    for the smartspim pipeline
    """
    return {
        "stitching": {
            "co_folder": "scratch",
            "channel": "Ex_488_Em_525",
            "resolution": {"x": "1.8", "y": "1.8", "z": "2.0"},
        },
        "registration": {"channels": ["Ex_488_Em_525"], "input_scale": "3"},
        # 'segmentation': {
        #     'channels': ['Ex_488_Em_525'],
        #     'input_scale': '0',
        #     'chunksize': '500',
        # },
    }


def build_pipeline_config(provided_config: dict, default_config: dict) -> dict:
    """
    Converts user input to pipeline input

    Parameters
    ----------
    provided_config: dict
        Dictionary provided from the processing_manifest.json

    default_config: dict
        Default configuration for each of the steps

    Returns
    ----------
    dict
        Pipeline configuration
    """

    new_config = default_config.copy()

    stitching_config = file_utils.helper_validate_key_dict(
        provided_config, "stitching"
    )

    cell_segmentation_channels = file_utils.helper_validate_key_dict(
        provided_config, "cell_segmentation_channels"
    )

    ccf_registration_channels = file_utils.helper_validate_key_dict(
        provided_config, "ccf_registration_channels"
    )

    # Setting stitching channel
    if stitching_config is not None:
        new_config["stitching"]["channel"] = stitching_config["channel"]
        new_config["stitching"]["resolution"]["x"] = str(
            stitching_config["resolution"]["x"]
        )
        new_config["stitching"]["resolution"]["y"] = str(
            stitching_config["resolution"]["y"]
        )
        new_config["stitching"]["resolution"]["z"] = str(
            stitching_config["resolution"]["z"]
        )

    else:
        default_ch = new_config["stitching"]["channel"]
        x_res = new_config["stitching"]["resolution"]["x"]
        y_res = new_config["stitching"]["resolution"]["y"]
        z_res = new_config["stitching"]["resolution"]["z"]

        res = f"x={x_res} um, y={y_res} um and z={z_res} um"
        logger.info(
            f"Using default stitching channel {default_ch} and resolution {res}"
        )

    # Setting CCF registration
    if ccf_registration_channels is not None:
        new_config["registration"]["channels"] = ccf_registration_channels

    else:
        default_ch = new_config["registration"]["channels"]
        logger.info(f"Using default registration channel {default_ch}")

    # Setting cell segmentation channels
    if cell_segmentation_channels is not None:
        new_config["segmentation"] = {"input_scale": "0", "chunksize": "500"}
        new_config["segmentation"]["channels"] = cell_segmentation_channels

    else:
        logger.info(f"No segmentation!")

    return new_config


def get_upload_datasets(
    dataset_folder: PathLike,
    config_path: PathLike,
    info_manager_path: Optional[PathLike] = None,
) -> List:
    """
    This function gets the datasets and
    classifies them in pending, uploaded
    and warning.

    - Pending: Datasets that need to be uploaded
    - Uploading: Datasets that are currently being uploaded
    - Uploaded: Datasets that are in the cloud
    - Warning: Datasets that are not ready to be
    uploaded

    Parameters
    ------------------------
    dataset_folder: PathLike
        Path where the smartspim datasets are located.

    config_path: PathLike
        Path inside each dataset where the processing_manifest.json
        is located

    Returns
    ------------------------
    List[dict]
        List with the pending datasets.
        This list contains the path and
        smartspim pipeline configuration
        for the datasets
    """

    # Lists for classifying datasets
    pending_datasets = []
    uploading_datasets = []
    uploaded_datasets = []
    warning_datasets = []

    # List of pending datasets with pipeline configuration
    # built just like code ocean accepts it
    pending_datasets_config = []
    default_pipeline_config = get_smartspim_default_config()

    # Get smartspim datasets that match data conventions
    dataset_folder = Path(dataset_folder)
    smartspim_datasets = SmartSPIMReader.read_smartspim_folders(dataset_folder)

    # Checking status
    for dataset in smartspim_datasets:
        dataset_path = dataset_folder.joinpath(dataset)

        # Reading config json
        dataset_config = file_utils.read_json_as_dict(
            dataset_path.joinpath(config_path)
        )

        dataset_path = str(dataset_path)
        dataset_status = file_utils.helper_validate_key_dict(
            dataset_config, "dataset_status"
        )

        if dataset_status is None:
            warning_datasets.append(dataset_path)
            continue

        # Validating status
        dataset_status = dataset_status.casefold()

        # Getting pipeline configuration to dataset
        pipeline_processing = file_utils.helper_validate_key_dict(
            dataset_config, "pipeline_processing"
        )

        if dataset_status == "pending" and pipeline_processing is not None:
            # Datasets to upload
            pending_datasets.append(dataset_path)
            pending_datasets_config.append(
                {
                    "path": dataset_path,
                    "pipeline_processing": build_pipeline_config(
                        pipeline_processing, default_pipeline_config
                    ),
                }
            )

        elif dataset_status == "uploading":
            # Datasets that are currently being uploaded
            uploading_datasets.append(dataset_path)

        # Using in instead of == since I add the upload time and s3 path
        elif "uploaded" in dataset_status:
            # Uploaded datasets
            uploaded_datasets.append(dataset_path)

        else:
            # Datasets that have issues
            warning_datasets.append(dataset_path)

    info_manager = {
        "generated_date": datetime.now().strftime("%d/%m/%Y %H:%M:%S"),
        "uploaded": uploaded_datasets,
        "uploading": uploading_datasets,
        "pending": pending_datasets,
        "warning": warning_datasets,
    }

    info_manager_path = (
        info_manager_path
        if info_manager_path is not None
        else dataset_folder.joinpath("status_smartspim_datasets.yaml")
    )
    file_utils.write_dict_to_yaml(info_manager, info_manager_path)

    return pending_datasets_config


def pre_upload_smartspim(
    dataset_path: PathLike, processing_manifest_path: PathLike
):
    """
    Steps that need to be performed
    before uploading data to the cloud

    Parameters
    ----------
    dataset_path: PathLike
        Path where the dataset is located

    processing_manifest_path: PathLike
        Path where the manifest json
        is stored in the data conventions
    """

    dataset_config_path = Path(dataset_path).joinpath(processing_manifest_path)
    msg = f"uploading"

    file_utils.update_json_key(
        json_path=dataset_config_path, key="dataset_status", new_value=msg
    )
    logger.info(f"Updating state to 'uploading'")


def provide_folder_permissions(
    root_folder: PathLike, paths: list, permissions: Optional[str] = "755"
):
    """
    Provides 755 permission in the folder.
    This fixes a known bug in the VAST system
    where folder are created with 000 permissions.

    Parameters
    -----------
    root_folder: PathLike
        Root folder where the datasets are located

    paths: List[PathLike]
        List with the paths to the folder that
        the permissions will be updated

    permissions: Optional[str]
        Permission that will be given via chmod

    """

    oct_permissions = int(permissions, 8)

    def helper_update_permissions(list_folders: list):
        """
        Helper function to update permissions
        of a list of folders

        Parameters
        -----------
        list_folders: list
            List of strings with the paths to
            the folders to update
        """
        for list_folder in list_folders:
            os.chmod(list_folder, oct_permissions)

    for path in paths:
        dataset_path = f"{root_folder}/{path}"
        curr_permissions = oct(os.stat(dataset_path).st_mode)[-3:]

        if curr_permissions != permissions:
            logger.info(
                f"Updating permissions for {dataset_path} from {curr_permissions} to {permissions}"
            )
            os.chmod(dataset_path, oct_permissions)

            # Updating first level
            first_level_folders = glob.glob(f"{dataset_path}/*/")
            second_level_folders = glob.glob(f"{dataset_path}/*/*/")
            third_level_folders = glob.glob(f"{dataset_path}/*/*/*/")

            helper_update_permissions(first_level_folders)
            helper_update_permissions(second_level_folders)
            helper_update_permissions(third_level_folders)


def main():
    """
    Main to execute the smartspim job
    """
    config_param = ArgSchemaParser(schema_type=ConfigFile)

    SUBMIT_HPC_PATH = (
        Path(os.path.dirname(os.path.realpath(__file__))) / "cluster/submit.py"
    )
    S3_UPLOAD_PATH = (
        Path(os.path.dirname(os.path.realpath(__file__))) / "s3_upload.py"
    )

    # Getting config file
    config_file_path = config_param.args["config_file"]
    config = get_default_config(config_file_path)

    # Code ocean credentials
    os.environ["CODEOCEAN_CREDENTIALS_PATH"] = config[
        "codeocean_credentials_path"
    ]

    # Exiftool path for image validation
    os.environ["PATH"] = f"{config['exiftool_path']}:{os.environ['PATH']}"

    # Get raw smartspim folders with processing_manifest.json
    root_folder = Path(config["root_folder"])

    (
        raw_datasets_ready,
        raw_datasets_rejected,
    ) = get_smartspim_folders_with_file(
        root_folder=root_folder, search_file="processing_manifest.json"
    )

    provide_folder_permissions(
        root_folder=root_folder, paths=raw_datasets_ready, permissions="755"
    )

    logger.warning(f"Raw datasets rejected: {raw_datasets_rejected}")

    new_dataset_paths, ready_datasets = organize_datasets(
        root_folder,
        raw_datasets_ready,
        config["metadata_service_domain"],
    )

    processing_manifest_path = "derivatives/processing_manifest.json"
    # List of datasets with its pipeline configuration
    pending_datasets_config = get_upload_datasets(
        dataset_folder=root_folder,
        config_path=processing_manifest_path,  # Pointing to this folder due to data conventions
        info_manager_path=config["info_manager_path"],
    )

    # Uploading datasets using the HPC
    sys.argv = [""]

    if config["dest_data_dir"] == None:
        config["dest_data_dir"] = ""

    mod = ArgSchemaParser(input_data=config, schema_type=CopyDatasets)
    args = mod.args

    # Getting params
    s3_bucket = args["s3_bucket"]
    nthreads = args["nthreads"]
    co_capsule_id = args["co_capsule_id"]
    dest_data_dir = Path(args["dest_data_dir"])

    logger.info(f"Uploading {pending_datasets_config}")

    for dataset in pending_datasets_config:
        # Adding the CO capsule to the dataset config
        dataset["co_capsule_id"] = co_capsule_id

        dataset_path = dataset["path"]
        dataset_name = Path(dataset_path).stem

        if os.path.isdir(dataset_path):
            dataset_dest_path = dest_data_dir.joinpath(dataset_name)

            if config["transfer_type"]["type"] == "HPC":
                dataset_dumped = json.dumps(dataset).replace('"', "[token]")
                cmd = f"""python {SUBMIT_HPC_PATH} generate-and-launch-run --job_cmd='python {S3_UPLOAD_PATH} \
                --input={dataset_path} --bucket={s3_bucket} --s3_path={dataset_name} --recursive --cluster \
                --trigger_code_ocean --pipeline_config="{dataset_dumped}" --type_spim="smartspim"' \
                --run_parent_dir='{config['transfer_type']['logs_folder']}' \
                --conda_activate='/home/{config['transfer_type']['hpc_account']}/anaconda3/bin/activate' \
                --conda_env='{config['transfer_type']['conda_env']}' --queue='{config['transfer_type']['hpc_queue']}' \
                --ntasks_per_node={config['transfer_type']['tasks_per_node']} \
                --nodes={config['transfer_type']['nodes']}  --cpus_per_task={config['transfer_type']['cpus_per_task']} \
                --mem_per_cpu={config['transfer_type']['mem_per_cpu']} --walltime='{config['transfer_type']['walltime']}' \
                --mail_user='{config['transfer_type']['mail_user']}' """

                # Setting dataset_status as 'uploading'
                pre_upload_smartspim(dataset_path, processing_manifest_path)

                # HPC run
                logger.info(f"Uploading dataset: {dataset_name}")
                for out in file_utils.execute_command(cmd):
                    logger.info(out)

                # Error with slurm logs directory
                time.sleep(30)
            else:
                # Local

                sys.argv = [
                    "",
                    f"--input={dataset_path}",
                    f"--bucket={s3_bucket}",
                    f"--s3_path={dataset_dest_path}",
                    f"--nthreads={nthreads}",
                    "--trigger_code_ocean",
                    f"--pipeline_config={dataset}",
                    "--recursive",
                ]

                # Local
                s3_upload.main()

                now_datetime = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
                dataset_config_path = Path(dataset_path).joinpath(
                    processing_manifest_path
                )
                msg = f"uploaded - Upload time: {now_datetime} - Bucket: {s3_bucket}"

                file_utils.update_json_key(
                    json_path=dataset_config_path,
                    key="dataset_status",
                    new_value=msg,
                )

        else:
            logger.warning(f"Path {dataset_path} does not exist. Ignoring...")

    pending_datasets = get_upload_datasets(
        dataset_folder=root_folder,
        config_path=processing_manifest_path,  # Pointing to this folder due to data conventions
        info_manager_path=config["info_manager_path"],
    )


if __name__ == "__main__":
    main()



"""Module to define and potentially run a BasicJob."""

import glob
import json
import logging
import os
import shutil
import sys
import tempfile
from datetime import datetime, timezone
from enum import Enum
from importlib import import_module
from pathlib import Path

import boto3
from aind_codeocean_api.codeocean import CodeOceanClient
from aind_data_schema.data_description import Modality

from aind_data_transfer import __version__
from aind_data_transfer.config_loader.base_config import BasicUploadJobConfigs
from aind_data_transfer.transformations.generic_compressors import (
    VideoCompressor,
    ZipCompressor,
)
from aind_data_transfer.transformations.metadata_creation import (
    ProceduresMetadata,
    ProcessingMetadata,
    RawDataDescriptionMetadata,
    SubjectMetadata,
)
from aind_data_transfer.util.s3_utils import upload_to_s3


# It might make more sense to move this class into different repo
class JobTypes(Enum):
    REGISTER_DATA = "register_data"
    RUN_GENERIC_PIPELINE = "run_generic_pipeline"
    TEST = "test"


class BasicJob:
    """Class that defines a basic upload job."""

    def __init__(self, job_configs: BasicUploadJobConfigs):
        """Init with job_configs"""
        self.job_configs = job_configs
        self._instance_logger = (
            logging.getLogger(__name__)
            .getChild(self.__class__.__name__)
            .getChild(str(id(self)))
        )
        self._instance_logger.setLevel(job_configs.log_level)

    def _test_upload(self, temp_dir: Path):
        """Run upload command on empty directory to see if user has permissions
        and aws cli installed.
        """
        upload_to_s3(
            directory_to_upload=temp_dir,
            s3_bucket=self.job_configs.s3_bucket,
            s3_prefix=self.job_configs.s3_prefix,
        )
        return None

    def _check_if_s3_location_exists(self):
        """Check if the s3 bucket and prefix already exists. If so, raise an
        error."""
        s3_client = boto3.client("s3")
        try:
            results = s3_client.list_objects_v2(
                Bucket=self.job_configs.s3_bucket,
                Prefix=self.job_configs.s3_prefix,
                MaxKeys=1,
            )
        finally:
            s3_client.close()
        if (
            results["KeyCount"] != 0
            and self.job_configs.force_cloud_sync is False
        ):
            raise FileExistsError(
                f"S3 path s3://{self.job_configs.s3_bucket}/"
                f"{self.job_configs.s3_prefix} already exists. Please consult "
                f"your admin for help removing old folder if desired."
            )
        elif (
            results["KeyCount"] != 0
            and self.job_configs.force_cloud_sync is True
        ):
            logging.warning(
                f"S3 path s3://{self.job_configs.s3_bucket}/"
                f"{self.job_configs.s3_prefix} already exists. Will force "
                f"syncing of local directory"
            )
        return None

    def _compress_raw_data(self, temp_dir: Path) -> None:
        """If compress data is set to False, the data will be uploaded to s3.
        Otherwise, it will be zipped, stored in temp_dir, and uploaded
        later."""

        if self.job_configs.behavior_dir is not None:
            behavior_dir_excluded = self.job_configs.behavior_dir / "*"
        else:
            behavior_dir_excluded = None

        for modality_config in self.job_configs.modalities:
            self._instance_logger.info(
                f"Starting to process {modality_config.source}"
            )
            if not modality_config.compress_raw_data:
                dst_dir = temp_dir / modality_config.default_output_folder_name
                shutil.copytree(
                    modality_config.source,
                    dst_dir,
                    ignore=behavior_dir_excluded,
                )
            else:
                self._instance_logger.info(
                    f"Compressing data folder: {modality_config.source}"
                )
                if modality_config.modality == Modality.ECEPHYS:
                    # Would prefer to not have imports here, but we have
                    # conditional dependencies
                    EcephysCompressionParameters = getattr(
                        import_module(
                            "aind_data_transfer.transformations."
                            "ephys_compressors"
                        ),
                        "EcephysCompressionParameters",
                    )
                    EphysCompressors = getattr(
                        import_module(
                            "aind_data_transfer.transformations."
                            "ephys_compressors"
                        ),
                        "EphysCompressors",
                    )

                    compression_params = EcephysCompressionParameters(
                        source=modality_config.source,
                        extra_configs=modality_config.extra_configs,
                    )
                    compression_params._number_id = modality_config.number_id
                    ecephys_compress_job = EphysCompressors(
                        job_configs=compression_params,
                        behavior_dir=self.job_configs.behavior_dir,
                        log_level=self.job_configs.log_level,
                    )
                    ecephys_compress_job.compress_raw_data(temp_dir=temp_dir)
                else:
                    zc = ZipCompressor(display_progress_bar=True)
                    compressed_data_folder_name = (
                        str(os.path.basename(modality_config.source)) + ".zip"
                    )
                    folder_path = (
                        temp_dir
                        / modality_config.default_output_folder_name
                        / compressed_data_folder_name
                    )
                    os.mkdir(folder_path.parent)
                    skip_dirs = (
                        None
                        if self.job_configs.behavior_dir is None
                        else [self.job_configs.behavior_dir]
                    )
                    zc.compress_dir(
                        input_dir=modality_config.source,
                        output_dir=folder_path,
                        skip_dirs=skip_dirs,
                    )
        return None

    def _compile_metadata(
        self, temp_dir: Path, process_start_time: datetime
    ) -> None:
        """Compile metadata files. Keeps the data in the temp_dir before
        uploading to s3."""

        # Get metadata from service
        subject_metadata = SubjectMetadata.from_service(
            domain=self.job_configs.metadata_service_domain,
            subject_id=self.job_configs.subject_id,
        )
        subject_file_name = temp_dir / subject_metadata.output_filename
        subject_metadata.write_to_json(subject_file_name)
        procedures_metadata = ProceduresMetadata.from_service(
            domain=self.job_configs.metadata_service_domain,
            subject_id=self.job_configs.subject_id,
        )
        procedures_file_name = temp_dir / procedures_metadata.output_filename
        procedures_metadata.write_to_json(procedures_file_name)
        modalities = [m.modality for m in self.job_configs.modalities]
        data_description_metadata = RawDataDescriptionMetadata.from_inputs(
            name=self.job_configs.s3_prefix,
            modality=modalities,
        )
        data_description_file_name = (
            temp_dir / data_description_metadata.output_filename
        )
        data_description_metadata.write_to_json(data_description_file_name)

        processing_end_time = datetime.now(timezone.utc)
        processing_metadata = ProcessingMetadata.from_modalities_configs(
            modality_configs=self.job_configs.modalities,
            start_date_time=process_start_time,
            end_date_time=processing_end_time,
            output_location=(
                f"s3://{self.job_configs.s3_bucket}/"
                f"{self.job_configs.s3_prefix}"
            ),
            code_url=self.job_configs.aind_data_transfer_repo_location,
            notes=None,
        )
        processing_file_name = temp_dir / processing_metadata.output_filename
        processing_metadata.write_to_json(processing_file_name)

        # Check user defined metadata
        if self.job_configs.metadata_dir:
            # Check only json files in user defined metadata folder
            file_paths = glob.glob(
                str(self.job_configs.metadata_dir / "*.json")
            )
            for file_path in file_paths:
                new_path = temp_dir / Path(file_path).name
                # Only overwrite service defined metadata if metadata_dir_force
                # is set to true
                if (
                    not os.path.exists(new_path)
                    or self.job_configs.metadata_dir_force
                ):
                    shutil.copyfile(file_path, new_path)
        return None

    def _encrypt_behavior_dir(self, temp_dir: Path) -> None:
        """Encrypt the data in the behavior directory. Keeps the data in the
        temp_dir before uploading to s3."""
        if self.job_configs.behavior_dir:
            encryption_key = (
                self.job_configs.video_encryption_password.get_secret_value()
            )
            video_compressor = VideoCompressor(encryption_key=encryption_key)
            new_behavior_dir = temp_dir / "behavior"
            # Copy data to a temp directory
            for root, dirs, files in os.walk(self.job_configs.behavior_dir):
                for file in files:
                    raw_file_path = os.path.join(root, file)
                    new_file_path = os.path.join(new_behavior_dir, file)
                    os.makedirs(os.path.dirname(new_file_path), exist_ok=True)
                    shutil.copy(raw_file_path, new_file_path)
            video_compressor.compress_all_videos_in_dir(new_behavior_dir)
        return None

    def _upload_to_s3(self, temp_dir: Path) -> None:
        """Upload the data to s3."""
        upload_to_s3(
            directory_to_upload=temp_dir,
            s3_bucket=self.job_configs.s3_bucket,
            s3_prefix=self.job_configs.s3_prefix,
            dryrun=self.job_configs.dry_run,
        )
        return None

    def _trigger_codeocean_pipeline(self):
        """Trigger the codeocean pipeline."""
        if self.job_configs.codeocean_process_capsule_id is not None:
            job_type = JobTypes.RUN_GENERIC_PIPELINE.value
        else:
            # Handle legacy way we set up parameters...
            job_type = self.job_configs.experiment_type.value
        trigger_capsule_params = {
            "trigger_codeocean_job": {
                "job_type": job_type,
                "modalities": (
                    [m.modality.name for m in self.job_configs.modalities]
                ),
                "capsule_id": self.job_configs.codeocean_trigger_capsule_id,
                "process_capsule_id": (
                    self.job_configs.codeocean_process_capsule_id
                ),
                "bucket": self.job_configs.s3_bucket,
                "prefix": self.job_configs.s3_prefix,
                "aind_data_transfer_version": __version__,
            }
        }
        co_client = CodeOceanClient(
            token=self.job_configs.codeocean_api_token.get_secret_value(),
            domain=self.job_configs.codeocean_domain,
        )
        if self.job_configs.dry_run:
            self._instance_logger.info(
                f"(dryrun) Would have ran capsule with params: "
                f"{trigger_capsule_params}"
            )
        else:
            response = co_client.run_capsule(
                capsule_id=self.job_configs.codeocean_trigger_capsule_id,
                data_assets=[],
                parameters=[json.dumps(trigger_capsule_params)],
            )
            self._instance_logger.info(
                f"Code Ocean Response: {response.json()}"
            )
        return None

    def run_job(self):
        """Runs the job. Creates a temp directory to compile the files before
        uploading."""
        process_start_time = datetime.now(timezone.utc)
        self._check_if_s3_location_exists()
        with tempfile.TemporaryDirectory(
            dir=self.job_configs.temp_directory
        ) as td:
            self._instance_logger.info("Checking write credentials...")
            self._test_upload(temp_dir=Path(td))
            self._instance_logger.info("Processing raw data...")
            self._compress_raw_data(temp_dir=Path(td))
            self._instance_logger.info("Checking for behavior directory...")
            self._encrypt_behavior_dir(temp_dir=Path(td))
            self._instance_logger.info("Compiling metadata...")
            self._compile_metadata(
                temp_dir=Path(td), process_start_time=process_start_time
            )
            self._instance_logger.info("Starting s3 upload...")
            self._upload_to_s3(temp_dir=Path(td))
            self._instance_logger.info("Initiating code ocean pipeline...")
            self._trigger_codeocean_pipeline()


if __name__ == "__main__":
    sys_args = sys.argv[1:]
    job_configs_from_main = BasicUploadJobConfigs.from_args(sys_args)
    job = BasicJob(job_configs=job_configs_from_main)
    job.run_job()



import unittest

import numpy as np

from aind_data_transfer.util import chunk_utils
from aind_data_transfer.util.chunk_utils import DimensionsError


class TestChunkUtils(unittest.TestCase):
    """Tests methods defined in chunk_utils class"""

    def test_ensure_array_5d(self):
        # Test padding if <5D
        a = np.zeros(shape=(1,) * 3, dtype=int)
        expected_arr = np.array([[[[[0]]]]])
        actual_arr = chunk_utils.ensure_array_5d(a)
        self.assertTrue(np.array_equal(expected_arr, actual_arr))

        # Test exception if >5D
        a = np.zeros(shape=(1,) * 6, dtype=int)
        with self.assertRaises(DimensionsError):
            chunk_utils.ensure_array_5d(a)

    def test_ensure_shape_5d(self):
        # Test padding if <5D
        s = (1,) * 3
        expected_shape = (1,) * 5
        actual_shape = chunk_utils.ensure_shape_5d(s)
        self.assertEqual(expected_shape, actual_shape)

        # Test exception if >5D
        s = (1,) * 6
        with self.assertRaises(DimensionsError):
            chunk_utils.ensure_shape_5d(s)

    def test_guess_chunks(self):
        data_shape = (2048, 2048, 2048)
        target_size = 128 * (1024**2)  # 128MB
        itemsize = 2

        expected_chunks = (16, 2048, 2048)
        actual_chunks = chunk_utils.guess_chunks(
            data_shape, target_size, itemsize, mode="z"
        )
        self.assertEqual(expected_chunks, actual_chunks)

        actual_chunks = chunk_utils.guess_chunks(
            data_shape, target_size, itemsize, mode="cycle"
        )
        expected_chunks = (256, 512, 512)
        self.assertEqual(expected_chunks, actual_chunks)

        expected_chunks = (407, 407, 407)
        actual_chunks = chunk_utils.guess_chunks(
            data_shape, target_size, itemsize, mode="iso"
        )
        self.assertEqual(expected_chunks, actual_chunks)

        # Test expected exceptions
        with self.assertRaises(ValueError):
            # shape dims == 0
            chunk_utils.guess_chunks((256, 256, 0), target_size, itemsize)

        with self.assertRaises(ValueError):
            # 0 target_size
            chunk_utils.guess_chunks(data_shape, 0, itemsize)

        with self.assertRaises(ValueError):
            # 0 item_size
            chunk_utils.guess_chunks(data_shape, target_size, 0)

    def test_expand_chunks(self):
        data_shape = (2048, 2048, 2048)
        chunks = (32, 32, 32)
        target_size = 128 * (1024**2)  # 128MB
        itemsize = 2

        expected_chunks = (416, 416, 416)
        actual_chunks = chunk_utils.expand_chunks(
            chunks, data_shape, target_size, itemsize, mode="iso"
        )
        self.assertEqual(expected_chunks, actual_chunks)

        expected_chunks = (512, 512, 256)
        actual_chunks = chunk_utils.expand_chunks(
            chunks, data_shape, target_size, itemsize, mode="cycle"
        )
        self.assertEqual(expected_chunks, actual_chunks)

        # Test expected exceptions
        with self.assertRaises(ValueError):
            # 0 chunk dims
            chunk_utils.expand_chunks(
                (256, 256, 0), data_shape, target_size, itemsize
            )

        with self.assertRaises(ValueError):
            # 0 shape dims
            chunk_utils.expand_chunks(
                chunks, (256, 256, 0), target_size, itemsize
            )

        with self.assertRaises(ValueError):
            # chunks larger than shape
            chunk_utils.expand_chunks(
                (128, 128, 129), (128, 128, 128), target_size, itemsize
            )

        with self.assertRaises(ValueError):
            # 0 target size
            chunk_utils.expand_chunks(chunks, data_shape, 0, itemsize)

        with self.assertRaises(ValueError):
            # 0 itemsize
            chunk_utils.expand_chunks(chunks, data_shape, target_size, 0)

    def test_closer_to_target(self):
        target_bytes = 128 * (1024**2)
        itemsize = 2

        s1 = (1024, 256, 256)
        s2 = (256, 256, 1024)
        closer = chunk_utils._closer_to_target(s1, s2, target_bytes, itemsize)
        self.assertEqual(closer, s2)

        s1 = (1024, 256, 256)
        s2 = (1024, 256, 257)
        closer = chunk_utils._closer_to_target(s1, s2, target_bytes, itemsize)
        self.assertEqual(closer, s1)

        s1 = (1024, 256, 255)
        s2 = (1024, 256, 256)
        closer = chunk_utils._closer_to_target(s1, s2, target_bytes, itemsize)
        self.assertEqual(closer, s2)

    def test_get_size(self):
        shape = (32, 256, 512)
        itemsize = 2

        expected_size = 8388608
        actual_size = chunk_utils._get_size(shape, itemsize)
        self.assertEqual(expected_size, actual_size)


if __name__ == "__main__":
    unittest.main()



import argparse
import itertools
import json
import logging
import os
import time
from datetime import datetime
from pathlib import PurePath

import numpy as np
from aind_codeocean_api.codeocean import CodeOceanClient
from aind_codeocean_api.credentials import CodeOceanCredentials
from s3transfer.constants import GB, MB

from aind_data_transfer.s3 import S3Uploader
from aind_data_transfer.util import file_utils
from aind_data_transfer.util.dask_utils import (
    get_client,
    log_dashboard_address,
)
from aind_data_transfer.util.file_utils import collect_filepaths

LOG_FMT = "%(asctime)s %(message)s"
LOG_DATE_FMT = "%Y-%m-%d %H:%M"

logging.basicConfig(format=LOG_FMT, datefmt=LOG_DATE_FMT)
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def chunk_files(input_dir, ntasks, recursive=True, exclude_dirs=None):
    filepaths = collect_filepaths(input_dir, recursive, exclude_dirs)

    filepaths_len = len(filepaths)

    if not filepaths_len:
        return None

    logger.info(f"Collected {filepaths_len} files")
    return np.array_split(filepaths, ntasks)


def upload_files_job(
    input_dir,
    files,
    bucket,
    s3_path,
    n_threads,
    target_throughput,
    part_size,
    timeout,
):
    uploader = S3Uploader(
        num_threads=n_threads,
        target_throughput=target_throughput,
        part_size=part_size,
        upload_timeout=timeout,
    )
    return uploader.upload_files(files, bucket, s3_path, root=input_dir)


def run_cluster_job(
    input_dir,
    bucket,
    s3_path,
    target_throughput,
    part_size,
    timeout,
    chunks_per_worker,
    recursive,
    exclude_dirs=None,
):
    client, ntasks = get_client(deployment="slurm")
    logger.info(f"Client has {ntasks} registered workers")

    chunked_files = chunk_files(
        input_dir, ntasks * chunks_per_worker, recursive, exclude_dirs
    )

    if chunked_files is None:
        logger.error(f"No files found!")
        client.close()
        return -1

    logger.info(
        f"Split files into {len(chunked_files)} chunks with "
        f"{len(chunked_files[0])} files each"
    )

    futures = []
    for chunk in chunked_files:
        futures.append(
            client.submit(
                upload_files_job,
                input_dir=input_dir,
                files=chunk,
                bucket=bucket,
                s3_path=s3_path,
                n_threads=1,
                target_throughput=target_throughput,
                part_size=part_size,
                timeout=timeout,
            )
        )
    failed_uploads = list(itertools.chain(*client.gather(futures)))
    n_failed_uploads = len(failed_uploads)
    logger.info(f"{n_failed_uploads} failed uploads:\n{failed_uploads}")
    client.close()

    return n_failed_uploads


def run_local_job(
    input_dir,
    bucket,
    s3_path,
    nthreads,
    target_throughput,
    part_size,
    timeout,
    recursive,
    exclude_dirs=None,
):
    uploader = S3Uploader(
        num_threads=nthreads,
        target_throughput=target_throughput,
        part_size=part_size,
        upload_timeout=timeout,
    )
    if os.path.isdir(input_dir):
        failed_uploads = uploader.upload_folder(
            input_dir, bucket, s3_path, recursive, exclude_dirs
        )
    elif os.path.isfile(input_dir):
        failed_uploads = uploader.upload_file(input_dir, bucket, s3_path)
    else:
        raise ValueError(
            f"Invalid value for --input: {input_dir} does not exist"
        )

    n_failed_uploads = len(failed_uploads)
    logger.info(f"{n_failed_uploads} failed uploads:\n{failed_uploads}")

    return n_failed_uploads


def get_smartspim_config(job_config: dict, pipeline_config: dict) -> dict:
    def get_key(job_cnf, key):
        if key in pipeline_config:
            job_cnf[key] = pipeline_config[key]

        return job_cnf

    for step in ["stitching", "registration", "segmentation"]:
        job_config = get_key(job_config, step)

    return job_config


def post_upload_smartspim(args: dict, s3_path: str, n_failed_uploads: int):
    """
    Post upload function for the smartspim
    datasets

    Parameters
    ----------
    args: dict
        Dictionary with the script
        arguments

    s3_path: str
        s3 path where the data was
        uploaded

    n_failed_uploads: int
        Number of failed uploads
    """

    if not n_failed_uploads:
        # Unpackaging args for easy code reading
        input_path = args.input
        bucket = args.bucket
        trigger_code_ocean = args.trigger_code_ocean
        pipeline_config = None

        if len(args.pipeline_config):
            pipeline_config = args.pipeline_config.replace("[token]", '"')
            pipeline_config = json.loads(pipeline_config)

        # Updating dataset_status in processing manifest
        now_datetime = datetime.now().strftime("%d/%m/%Y %H:%M:%S")

        processing_manifest_path = "derivatives/processing_manifest.json"
        dataset_config_path = PurePath(input_path).joinpath(
            processing_manifest_path
        )
        msg = f"uploaded - Upload time: {now_datetime} - Bucket: {bucket}"

        file_utils.update_json_key(
            json_path=dataset_config_path, key="dataset_status", new_value=msg
        )

        # Triggering code ocean if necessary
        if trigger_code_ocean and pipeline_config:
            logger.info(f"Triggering code ocean {pipeline_config}")
            asset_name = PurePath(s3_path).stem

            job_configs = {
                "trigger_codeocean_job": {
                    "job_type": "smartspim",
                    "bucket": bucket,
                    "prefix": asset_name,
                }
            }

            job_configs["trigger_codeocean_job"] = get_smartspim_config(
                job_configs["trigger_codeocean_job"],
                pipeline_config["pipeline_processing"],
            )

            capsule_id = pipeline_config["co_capsule_id"]

            try:
                co_cred = CodeOceanCredentials().credentials

                co_api = CodeOceanClient(
                    domain=co_cred["domain"], token=co_cred["token"]
                )

                run_response = co_api.run_capsule(
                    capsule_id=capsule_id,
                    data_assets=[],
                    parameters=[json.dumps(job_configs)],
                )
                logger.info(f"Run response: {run_response.json()}")

            except ValueError as err:
                logger.error(f"Error communicating with Code Ocean API {err}")

        else:
            logger.warning(
                "Code ocean was not triggered. If this is an error, check your parameters"
            )

    else:
        logger.error(
            f"n failed uploads: {n_failed_uploads}. Please, validate."
        )
        logger.error(f"Skipping code ocean execution!")


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-i",
        "--input",
        type=str,
        help="folder or file to upload",
    )
    parser.add_argument(
        "-b",
        "--bucket",
        type=str,
        help="s3 bucket",
    )
    parser.add_argument(
        "--s3_path",
        type=str,
        default=None,
        help='s3 path relative to bucket, e.g., "folder/data.h5"',
    )
    parser.add_argument(
        "--target_throughput",
        type=float,
        default=100 * GB / 8,
        help="target throughput (bytes)",
    )
    parser.add_argument(
        "--part_size",
        type=int,
        default=64 * MB,
        help="part size for s3 multipart uploads (bytes)",
    )
    parser.add_argument(
        "--timeout",
        type=float,
        default=None,
        help="per-file upload timeout (s). Default is None.",
    )
    parser.add_argument(
        "--cluster",
        default=False,
        action="store_true",
        help="run in cluster mode",
    )
    parser.add_argument(
        "--batch_num",
        type=int,
        default=3,
        help="number of tasks per job. "
        "Increase this if you run into worker memory issues",
    )
    parser.add_argument(
        "--nthreads",
        type=int,
        default=1,
        help="num threads to use if running locally",
    )
    parser.add_argument(
        "--recursive",
        default=False,
        action="store_true",
        help="upload a directory recursively",
    )
    parser.add_argument(
        "--exclude_dirs",
        type=str,
        nargs="+",
        default=None,
        help="directories to exclude from upload",
    )
    parser.add_argument(
        "--type_spim",
        type=str,
        default="",
        help="Type of SPIM dataset. e.g., SmartSPIM, ExASPIM, diSPIM...",
    )
    parser.add_argument(
        "--trigger_code_ocean",
        default=False,
        action="store_true",
        help="upload a directory recursively",
    )
    parser.add_argument(
        "--pipeline_config",
        type=str,
        default="",
        help="Configuration to pipeline in Code Ocean",
    )

    args = parser.parse_args()

    input_path = args.input
    bucket = args.bucket
    n_failed_uploads = -1
    trigger_code_ocean = args.trigger_code_ocean
    type_spim = args.type_spim.casefold()

    s3_path = args.s3_path
    if s3_path is None:
        s3_path = PurePath(args.input).name
    # remove leading slash since it will result in the structure
    # "bucket/'/'/s3_path" instead of "bucket/s3_path"
    s3_path = s3_path.strip("/")
    logger.info(f"Will upload to {args.bucket}/{s3_path}")

    t0 = time.time()

    # Uploading data
    if args.cluster:
        n_failed_uploads = run_cluster_job(
            input_dir=input_path,
            bucket=bucket,
            s3_path=s3_path,
            target_throughput=args.target_throughput,
            part_size=args.part_size,
            timeout=args.timeout,
            chunks_per_worker=args.batch_num,
            recursive=args.recursive,
            exclude_dirs=args.exclude_dirs,
        )
    else:
        n_failed_uploads = run_local_job(
            input_dir=input_path,
            bucket=bucket,
            s3_path=s3_path,
            nthreads=args.nthreads,
            target_throughput=args.target_throughput,
            part_size=args.part_size,
            timeout=args.timeout,
            recursive=args.recursive,
            exclude_dirs=args.exclude_dirs,
        )

    logger.info(f"Upload done. Took {time.time() - t0}s ")

    # Different post processing executions for spim data
    if type_spim == "smartspim":
        post_upload_smartspim(args, s3_path, n_failed_uploads)


if __name__ == "__main__":
    main()



"""
Imaging writers for the different data modalities
"""
import logging
import os
import re
from datetime import date, datetime, time
from pathlib import Path
from typing import Any, List, Tuple, Union

import chardet
from aind_data_schema import Funding, RawDataDescription, Subject
from aind_data_schema.data_description import (
    ExperimentType,
    Group,
    Institution,
    Modality,
)
from aind_data_schema.imaging import acquisition, tile
from aind_metadata_service.client import AindMetadataServiceClient

from aind_data_transfer.readers.imaging_readers import SmartSPIMReader
from aind_data_transfer.util import file_utils

PathLike = Union[str, Path]

logging.basicConfig(
    level=logging.DEBUG,
    format="%(asctime)s - %(levelname)s : %(message)s",
    datefmt="%Y-%m-%d %H:%M",
    handlers=[
        logging.StreamHandler(),
        # logging.FileHandler("test.log", "a"),
    ],
)
logging.disable("DEBUG")
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def digest_asi_line(line: str) -> datetime:
    """
    Scrape a datetime from a non-empty line, otherwise return None

    Parameters
    -----------
    line: str
        Line from the ASI file

    Returns
    -----------
    datetime
        A date that could be parsed from a string
    """

    if line.isspace():
        return None
    else:
        mdy, hms, ampm = line.split()[0:3]

    mdy = [int(i) for i in mdy.split(b"/")]
    ymd = [mdy[i] for i in [2, 0, 1]]

    hms = [int(i) for i in hms.split(b":")]
    if ampm == b"PM":
        hms[0] += 12
        if hms[0] == 24:
            hms[0] = 0

    ymdhms = ymd + hms

    dtime = datetime(*ymdhms)
    ymd = date(*ymd)
    hms = time(*hms)
    return dtime


def get_session_end(asi_file) -> datetime:
    """
    Work backward from the last line until there is a timestamp

    Parameters
    ------------
    asi_file: PathLike
        Path where the ASI metadata file is
        located

    Returns
    ------------
    Date when the session ended
    """

    with open(asi_file, "rb") as file:
        asi_mdata = file.readlines()

    idx = -1
    result = None
    while result is None:
        result = digest_asi_line(asi_mdata[idx])
        idx -= 1

    return result


def get_line_indices_metadata_file(lines: bytes) -> dict:
    """
    Get the index location in the text file
    where each block of metadata information
    starts

    Parameters
    -----------
    lines: bytes
        Bytes with the metadata information
        taken from the text file

    Returns
    -----------
    dict
        Dictionary with the indices of
        each block of metadata
    """
    indices = {
        "info_start": None,
        "wavelength_start": None,
        "tile_acquisition_start": None,
    }

    total_lines = len(lines)

    for idx_line in range(total_lines):
        splited_line = lines[idx_line].rstrip().split("\t")

        if "Obj" in splited_line:
            indices["info_start"] = idx_line

        if "Wavelength" in splited_line:
            indices["wavelength_start"] = idx_line

        if "Exposure" in splited_line:
            indices["tile_acquisition_start"] = idx_line
            break

    return indices


def get_session_config(lines: List[str]) -> dict:
    """
    Gets the session config

    Parameters
    -----------
    lines: List[str]
        List with the lines of the
        metadata file

    Returns
    -----------
    Dictionary with the parsed data
    """
    # Get objective magnification
    objective_regex = r"\d+\.?\d*"  # "(\d+\.?\d*)(x|X)" to get with X

    data = {}
    lines_splited = lines[1].rstrip().split("\t")

    # Getting first line metadata
    obj = lines_splited[0]
    v_res = float(lines_splited[1])
    um_per_pix = float(lines_splited[2])
    z_step = float(lines_splited[3])
    scanning = lines_splited[4]
    sampling = lines_splited[5]
    destripe = lines_splited[6]
    z_block = int(lines_splited[7])

    data["obj_name"] = obj
    data["obj_magnification"] = re.findall(objective_regex, obj)[0]
    data["v_res"] = v_res
    data["um/pix"] = um_per_pix
    data["z_step_um"] = z_step
    data["scanning"] = scanning
    data["sampling"] = sampling
    data["destripe"] = destripe
    data["z_block"] = z_block

    return data


def get_wavelength_config(lines: List[str], start_index: int, end_index: int):
    """
    Gets the wavelength config

    Parameters
    -----------
    lines: List[str]
        List with the lines of the
        metadata file

    start_index: int
        Index with the line where
        we'll parse the data

    end_index: int
        Index with the line where
        we'll finish

    Returns
    -----------
    Dictionary with the parsed data
    """
    wavelengths = {}
    for line in lines[start_index:end_index]:
        line_splited = line.rstrip().split("\t")

        try:
            wavelength = int(line_splited[0])
            power_l = float(line_splited[1])
            power_r = float(line_splited[2])

            wavelengths[wavelength] = {}
            wavelengths[wavelength]["power_left"] = power_l
            wavelengths[wavelength]["power_right"] = power_r

        except ValueError as err:
            logger.error(
                f"An error ocurred while parsing {line_splited} Traceback: {err}"
            )
    return wavelengths


def get_tile_info(lines: List[str], start_index: int):
    """
    Gets the tile config

    Parameters
    -----------
    lines: List[str]
        List with the lines of the
        metadata file

    start_index: int
        Index with the line where
        we'll parse the data

    Returns
    -----------
    Dictionary with the parsed data
    """

    data = {}
    ite = 0

    for line in lines[start_index:]:
        splited_line = line.rstrip().split("\t")
        line_elements = len(splited_line)

        if not line_elements or (line_elements == 1 and splited_line[0] == ""):
            # End of file when reading from bucket
            break

        name = f"t_{ite}"

        x = int(splited_line[0])
        y = int(splited_line[1])
        z = int(splited_line[2])
        wavelength = int(splited_line[3])
        side = int(splited_line[4])
        exposure = int(splited_line[5])
        skip = int(splited_line[6])

        data[name] = {
            "x": x,
            "y": y,
            "z": z,
            "wavelength": wavelength,
            "side": side,
            "exposure": exposure,
            "skip": skip,
        }
        ite += 1

    return data


def make_acq_tiles(metadata_dict: dict, filter_mapping: dict):
    """
    Makes metadata for the acquired tiles of
    the dataset

    Parameters
    -----------
    metadata_dict: dict
        Dictionary with the acquisition metadata
        coming from the microscope

    filter_mapping: dict
        Dictionary with the channel names

    Returns
    -----------
    List[tile.Translation3dTransform]
        List with the metadata for the tiles
    """

    channels = {}

    # List where the metadata of the acquired
    # tiles is stored
    tile_acquisitions = []

    filter_wheel_idx = 0
    for wavelength, value in metadata_dict["wavelength_config"].items():
        channel = tile.Channel(
            channel_name=wavelength,
            laser_wavelength=wavelength,
            laser_power=value["power_left"],
            filter_wheel_index=filter_wheel_idx,
        )
        filter_wheel_idx += 1

        channels[wavelength] = channel

    # Scale metadata
    scale = tile.Scale3dTransform(
        scale=[
            metadata_dict["session_config"]["um/pix"],  # X res
            metadata_dict["session_config"]["um/pix"],  # Y res
            metadata_dict["session_config"]["z_step_um"],  # Z res
        ]
    )

    for tile_key, tile_info in metadata_dict["tile_config"].items():
        tile_transform = tile.Translation3dTransform(
            translation=[
                tile_info["x"] / 10,
                tile_info["y"] / 10,
                tile_info["z"] / 10,
            ]
        )

        channel = channels[tile_info["wavelength"]]

        tile_acquisition = tile.AcquisitionTile(
            channel=channel,
            notes=(
                "\nLaser power is in percentage of total, it needs calibration"
            ),
            coordinate_transformations=[tile_transform, scale],
            file_name=f"Ex_{tile_info['wavelength']}_Em_{filter_mapping[tile_info['wavelength']]}/{tile_info['x']}/{tile_info['x']}_{tile_info['y']}/",
        )

        tile_acquisitions.append(tile_acquisition)

    return tile_acquisitions


class SmartSPIMWriter:
    """This class contains the methods to write smartspim data."""

    def __init__(self, dataset_paths: dict, metadata_domain: str):
        """
        Class constructor.

        Parameters
        ------------------------
        dataset_paths: dict
            Dictionary with the dataset paths

        metadata_domain: str
            Metadata domain

        """
        self.__dataset_paths = dataset_paths
        self.__regex_expressions = SmartSPIMReader.RegexPatterns
        self.__metadata_domain = metadata_domain

    @property
    def dataset_paths(self) -> dict:
        """
        Getter of dataset paths.

        Returns
        ------------------------
        dict
            Dictionary with the dataset paths

        """
        return self.__dataset_paths

    @dataset_paths.setter
    def dataset_paths(self, new_dataset_paths: dict) -> None:
        """
        Setter of dataset paths.

        Parameters
        ------------------------
        new_dataset_paths: dict
            Dictionary with the dataset paths

        """
        self.__dataset_paths = new_dataset_paths

    def create_dataset_convention(self, dataset_path: PathLike) -> Tuple:
        """
        Creates the dataset name based on the data convention.

        Parameters
        ------------------------
        dataset_path: PathLike
            Path where the dataset is located

        Returns
        ------------------------
        Tuple
            Tuple with the new dataset path based on
            the data conventions and mouse id
        """

        dataset_path = Path(dataset_path)

        dataset_name = dataset_path.stem

        try:
            date_str = re.search(
                self.__regex_expressions.capture_date_regex.value, dataset_name
            ).group(1)
            time_str = re.search(
                self.__regex_expressions.capture_time_regex.value, dataset_name
            ).group(1)[1:-1]
            mouse_id_str = re.search(
                self.__regex_expressions.capture_mouse_id.value, dataset_name
            ).group(1)[1:]
        except ValueError as err:
            raise ValueError(
                "It was not possible to capture date, time or mouse_id.", err
            )

        date_time_obj = datetime.strptime(
            date_str + time_str, "%Y%m%d%H_%M_%S"
        )

        date_fmt = "%Y-%m-%d"
        date_str = date_time_obj.strftime(date_fmt)

        time_fmt = "%H-%M-%S"
        time_str = date_time_obj.strftime(time_fmt)

        new_dataset_path = dataset_path.parent.joinpath(
            f"SmartSPIM_{mouse_id_str}_{date_str}_{time_str}"
        )

        parsed_data = {"mouse_id": mouse_id_str, "mouse_date": date_time_obj}

        return new_dataset_path, parsed_data

    def __create_data_description(
        self, parsed_data: dict, dataset_info: dict, output_path: PathLike
    ):
        """
        Creates the data description json.

        Parameters
        ------------------------
        parsed_data: dict
            Parsed data for the dataset. Contains
            mouse id and creation datetime

        dataset_info: dict
            Information for the dataset

        output_path: PathLike
            Path where the dataset is located

        """

        mouse_date = parsed_data["mouse_date"]

        # Validating data_description args
        institution = None

        # Getting the institution
        if "institution" not in dataset_info:
            raise ValueError("Please, provide the institution in the manifest")
        else:
            institution = dataset_info["institution"]

        funding_source = file_utils.helper_validate_key_dict(
            dictionary=dataset_info,
            key="funding_source",
            default_return=institution,
        )
        project_name = file_utils.helper_validate_key_dict(
            dictionary=dataset_info, key="project"
        )
        project_id = file_utils.helper_validate_key_dict(
            dictionary=dataset_info, key="project_id"
        )
        group = file_utils.helper_validate_key_dict(
            dictionary=dataset_info, key="group", default_return="MSMA"
        )

        # Creating data description
        data_description = RawDataDescription(
            modality=[Modality.SPIM],
            subject_id=parsed_data["mouse_id"],
            creation_date=date(
                mouse_date.year, mouse_date.month, mouse_date.day
            ),
            creation_time=time(
                mouse_date.hour, mouse_date.minute, mouse_date.second
            ),
            institution=institution,
            group=group,
            project_name=project_name,
            project_id=project_id,
            funding_source=[Funding(funder=funding_source)],
            experiment_type=ExperimentType.SMARTSPIM,
        )

        data_description_path = str(
            output_path.joinpath("data_description.json")
        )

        with open(data_description_path, "w") as f:
            f.write(data_description.json(indent=3))

    def __create_subject(self, mouse_id: str, output_path: PathLike):
        """
        Creates the data description json.

        Parameters
        ------------------------
        mouse_id: str
            Mouse id for the dataset

        output_path: PathLike
            Path where the dataset is located
        """

        client = AindMetadataServiceClient(self.__metadata_domain)

        response = client.get_subject(mouse_id)

        if response.status_code == 200:
            data = response.json()["data"]

            subject = Subject(
                species=data["species"],
                subject_id=data["subject_id"],
                sex=data["sex"],
                date_of_birth=data["date_of_birth"],
                genotype=data["genotype"],
                mgi_allele_ids=data["mgi_allele_ids"],
                background_strain=data["background_strain"],
                source=data["source"],
                rrid=data["rrid"],
                restrictions=data["restrictions"],
                breeding_group=data["breeding_group"],
                maternal_id=data["maternal_id"],
                maternal_genotype=data["maternal_genotype"],
                paternal_id=data["paternal_id"],
                paternal_genotype=data["paternal_genotype"],
                wellness_reports=data["wellness_reports"],
                notes=data["notes"],
            )

            subject_path = str(output_path.joinpath("subject.json"))

            with open(subject_path, "w") as f:
                f.write(subject.json(indent=3))

        else:
            logger.error(
                f"Mouse {mouse_id} does not have subject information - res status: {response.status_code}"
            )

    def __get_excitation_emission_waves(self, dataset_path: PathLike) -> dict:
        """
        Gets the excitation and emission waves for
        the existing channels within a dataset

        Parameters
        ------------
        dataset_path: PathLike
            Path where the channels of the dataset
            are stored

        Returns
        ------------
        dict
            Dictionary with the excitation
            and emission waves
        """
        regex_channels = self.__regex_expressions.regex_channels.value
        excitation_emission_channels = {}

        elements = [
            element
            for element in os.listdir(dataset_path)
            if re.match(regex_channels, element)
        ]

        for channel in elements:
            channel = channel.replace("Em_", "").replace("Ex_", "")
            splitted = channel.split("_")
            excitation_emission_channels[int(splitted[0])] = int(splitted[1])

        return excitation_emission_channels

    def __create_acquisition(
        self,
        parsed_data: dict,
        dataset_info: dict,
        original_dataset_path: PathLike,
        output_path: PathLike,
    ):
        """
        Creates the data description json.

        Parameters
        ------------------------
        parsed_data: dict
            Parsed data for the dataset. Contains
            mouse id and creation datetime

        dataset_info: dict
            Information for the dataset

        output_path: PathLike
            Path where the dataset is located

        """

        asi_file = original_dataset_path.joinpath("ASI_logging.txt")
        mdata_file = original_dataset_path.joinpath("metadata.txt")
        filter_mapping = self.__get_excitation_emission_waves(
            original_dataset_path
        )

        with open(mdata_file, "rb") as f:
            lc_mdata_result = chardet.detect(f.read())

        with open(mdata_file, "r", encoding=lc_mdata_result["encoding"]) as f:
            lc_mdata = f.readlines()

        # Get information where starts each metadata block
        # in the metadata file
        line_indices = get_line_indices_metadata_file(lc_mdata)

        # Parse first section
        session_config = get_session_config(lines=lc_mdata)

        # Getting wavelengths
        wavelength_config = get_wavelength_config(
            lines=lc_mdata,
            start_index=line_indices["wavelength_start"] + 1,
            end_index=line_indices["tile_acquisition_start"],
        )

        # Getting tile info
        tile_config = get_tile_info(
            lines=lc_mdata,
            start_index=line_indices["tile_acquisition_start"] + 1,
        )

        metadata_dict = {
            "session_config": session_config,
            "wavelength_config": wavelength_config,
            "tile_config": tile_config,
        }

        session_end_time = get_session_end(asi_file)

        # Validating data in config
        instrument_id = file_utils.helper_validate_key_dict(
            dictionary=dataset_info, key="instrument_id"
        )
        experimenter_full_name = file_utils.helper_validate_key_dict(
            dictionary=dataset_info, key="experimenter"
        )
        local_storage_directory = file_utils.helper_validate_key_dict(
            dictionary=dataset_info, key="local_storage"
        )
        chamber_immersion_medium = file_utils.helper_validate_key_dict(
            dictionary=dataset_info["chamber_immersion"], key="medium"
        )
        chamber_immersion_ri = file_utils.helper_validate_key_dict(
            dictionary=dataset_info["chamber_immersion"],
            key="refractive_index",
        )
        sample_immersion_medium = file_utils.helper_validate_key_dict(
            dictionary=dataset_info["sample_immersion"], key="medium"
        )
        sample_immersion_ri = file_utils.helper_validate_key_dict(
            dictionary=dataset_info["sample_immersion"], key="refractive_index"
        )

        # Giving the specific error of what is missing
        if instrument_id is None:
            raise ValueError("Instrument id not provided in manifest")

        if experimenter_full_name is None:
            raise ValueError("Experimenter full name not provided in manifest")

        if chamber_immersion_medium is None:
            raise ValueError(
                "Chamber immersion medium not provided in manifest"
            )

        if chamber_immersion_ri is None:
            raise ValueError("Chamber immersion ri not provided in manifest")

        acquisition_model = acquisition.Acquisition(
            specimen_id="",
            instrument_id=instrument_id,
            experimenter_full_name=experimenter_full_name,
            subject_id=parsed_data["mouse_id"],
            session_start_time=parsed_data["mouse_date"],
            session_end_time=session_end_time,
            local_storage_directory=local_storage_directory,
            external_storage_directory="",
            chamber_immersion=acquisition.Immersion(
                medium=chamber_immersion_medium,
                refractive_index=chamber_immersion_ri,
            ),
            sample_immersion=acquisition.Immersion(
                medium=sample_immersion_medium,
                refractive_index=sample_immersion_ri,
            ),
            axes=[
                acquisition.Axis(
                    name="X",
                    dimension=2,
                    direction="Left_to_right",
                ),
                acquisition.Axis(
                    name="Y", dimension=1, direction="Posterior_to_anterior"
                ),
                acquisition.Axis(
                    name="Z", dimension=0, direction="Superior_to_inferior"
                ),
            ],
            tiles=make_acq_tiles(
                metadata_dict=metadata_dict, filter_mapping=filter_mapping
            ),
        )

        acquisition_path = str(output_path.joinpath("acquisition.json"))

        with open(acquisition_path, "w") as f:
            f.write(acquisition_model.json(indent=3))

    def __create_smartspim_metadata(
        self,
        parsed_data: dict,
        dataset_info: dict,
        original_dataset_path: PathLike,
        output_path: PathLike,
    ) -> None:
        """
        Creates the data description json.

        Parameters
        ------------------------
        parsed_data: config
            Dictionary with the data from the dataset name.
            It includes mouse id, creation date and time for the dataset

        dataset_info: dict
            Information for the dataset

        dataset_path: PathLike
            Path where the dataset is located
        """

        output_path = Path(output_path)

        # Creates the data description json
        if "data_description" in dataset_info:
            self.__create_data_description(
                parsed_data, dataset_info["data_description"], output_path
            )
        else:
            logger.error(
                f"data_description.json was not created for {parsed_data['mouse_id']}. Add it to the YAML configuration."
            )

        # Creates the subject metadata json
        self.__create_subject(parsed_data["mouse_id"], output_path)

        # Creates the acquisition json
        if "acquisition" in dataset_info:
            try:
                self.__create_acquisition(
                    parsed_data,
                    dataset_info["acquisition"],
                    original_dataset_path,
                    output_path,
                )
            except ValueError:
                logger.error("Error creating acquisition schema")
        else:
            logger.error(
                f"acquisition.json was not created for {parsed_data['mouse_id']}. Add it to the YAML configuration."
            )

    def prepare_datasets(
        self, mode: str = "move", delete_empty: bool = True
    ) -> Tuple:
        """
        Prepares the smartspim folder structure
        based on the data conventions.

        Parameters
        ------------------------
        mode: str
            Preparation mode. Move to move data
            and copy to copy it.

        delete_empty: bool
            Deletes the original folder of the data
            if it's empty.

        Returns
        ------------------------
        Tuple
            Tuple with the new dataset paths and
            the ignored datasets
        """

        new_dataset_paths = []
        ignored_datasets = []

        for dataset_info in self.__dataset_paths:
            dataset_path = dataset_info["path"]

            if os.path.isdir(dataset_path):
                logger.info(f"\nOrganizing: {dataset_path}")
                (
                    new_dataset_path,
                    parsed_data,
                ) = self.create_dataset_convention(dataset_path)
                derivatives_path = new_dataset_path.joinpath("derivatives")
                smartspim_channels_path = new_dataset_path.joinpath(
                    "SmartSPIM"
                )

                if not os.path.isdir(new_dataset_path):
                    file_utils.create_folder(derivatives_path, True)
                    file_utils.create_folder(smartspim_channels_path, True)

                    # Temporary while we are able to find a way to get all metadata from datasets

                    # Create smartspim metadata
                    self.__create_smartspim_metadata(
                        parsed_data=parsed_data,
                        dataset_info=dataset_info,
                        original_dataset_path=dataset_path,
                        output_path=new_dataset_path,
                    )

                    # Moving channels
                    file_utils.move_folders_or_files(
                        dataset_path,
                        new_dataset_path.joinpath("SmartSPIM"),
                        self.__regex_expressions.regex_channels.value,
                        mode=mode,
                    )

                    # Moving maximal intensity projections per channel
                    file_utils.move_folders_or_files(
                        dataset_path,
                        new_dataset_path.joinpath("derivatives"),
                        self.__regex_expressions.regex_channels_MIP.value,
                        mode=mode,
                    )

                    # Moving metadata files
                    file_utils.move_folders_or_files(
                        dataset_path,
                        new_dataset_path.joinpath("derivatives"),
                        self.__regex_expressions.regex_files.value,
                        mode=mode,
                    )

                    new_dataset_paths.append(new_dataset_path)

                else:
                    ignored_datasets.append(new_dataset_path)
                    logger.warning(
                        f"[!] {new_dataset_path} already exists, please check it. Ignoring..."
                    )

                # We delete the folder if it does not have any files
                elements_in_dataset = os.listdir(dataset_path)

                if delete_empty and not len(elements_in_dataset):
                    # Delete only if folder is empty
                    file_utils.delete_folder(dataset_path)

            else:
                logger.warning(
                    f"[!] Path not found for dataset located in: {dataset_path}. Ignoring..."
                )

        return new_dataset_paths, ignored_datasets



"""Loads Imaging job configurations"""
import argparse
from pathlib import Path

import yaml
from numcodecs import Blosc

from aind_data_transfer.util.file_utils import is_cloud_url, parse_cloud_url


class ImagingJobConfigurationLoader:
    def load_configs(self, sys_args):
        """Load yaml config at conf_src Path as python dict"""
        parser = argparse.ArgumentParser()
        parser.add_argument(
            "-c", "--conf-file-location", required=True, type=str
        )
        parser.add_argument(
            "-r", "--raw-data-source", required=False, type=str
        )
        args = parser.parse_args(sys_args)
        conf_src = args.conf_file_location
        with open(conf_src) as f:
            config = yaml.load(f, Loader=yaml.SafeLoader)
        if args.raw_data_source is not None:
            config["endpoints"]["raw_data_dir"] = args.raw_data_source
        self.__resolve_endpoints(config)
        self.__parse_compressor_configs(config)
        return config

    @staticmethod
    def __resolve_endpoints(configs):
        """
        If the destination folder is a cloud bucket without prefix, a prefix
        will be created using the acquisition directory name.
        Args:
            configs (dic): Configurations

        Returns:
            None, modifies the base configs in place
        """

        dest_data_dir = configs["endpoints"]["dest_data_dir"]
        if is_cloud_url(dest_data_dir):
            provider, bucket, prefix = parse_cloud_url(dest_data_dir)
            if prefix == "":
                prefix = Path(configs["endpoints"]["raw_data_dir"]).name
                dest_data_dir = dest_data_dir.strip("/") + "/" + prefix
                configs["endpoints"]["dest_data_dir"] = dest_data_dir

    @staticmethod
    def __parse_compressor_configs(configs):
        """Util method to map a string to class attribute"""
        try:
            compressor_name = configs["transcode_job"]["compressor"][
                "compressor_name"
            ]
        except KeyError:
            compressor_name = None
        try:
            compressor_kwargs = configs["transcode_job"]["compressor"][
                "kwargs"
            ]
        except KeyError:
            compressor_kwargs = None
        if (
            compressor_name
            and compressor_name == Blosc.codec_id
            and compressor_kwargs
            and "shuffle" in compressor_kwargs
        ):
            shuffle_str = configs["transcode_job"]["compressor"]["kwargs"][
                "shuffle"
            ]
            shuffle_val = getattr(Blosc, shuffle_str)
            configs["transcode_job"]["compressor"]["kwargs"][
                "shuffle"
            ] = shuffle_val



"""Test module for generic upload job."""

import json
import os
import unittest
from pathlib import Path
from unittest.mock import MagicMock, call, patch

from aind_data_schema.data_description import ExperimentType, Modality

from aind_data_transfer.config_loader.base_config import ModalityConfigs
from aind_data_transfer.jobs.s3_upload_job import GenericS3UploadJobList

TEST_DIR = Path(os.path.dirname(os.path.realpath(__file__)))
CONFIGS_DIR = TEST_DIR / "resources" / "test_configs"
METADATA_DIR = TEST_DIR / "resources" / "test_metadata"


class TestGenericS3UploadJobList(unittest.TestCase):
    """Unit tests for methods in GenericS3UploadJobs class."""

    EXAMPLE_PARAM_STORE_RESPONSE = json.dumps(
        {
            "codeocean_domain": "some_domain",
            "codeocean_trigger_capsule_id": "some_capsule_id",
            "metadata_service_domain": "some_ms_domain",
            "aind_data_transfer_repo_location": "some_dtr_location",
            "video_encryption_password_path": "/path/to/password",
            "codeocean_api_token_path": "/path/to/token",
        }
    )

    PATH_TO_EXAMPLE_CSV_FILE = (
        Path(os.path.dirname(os.path.realpath(__file__)))
        / "resources"
        / "test_configs"
        / "jobs_list.csv"
    )

    PATH_TO_EXAMPLE_CSV_FILE2 = (
        Path(os.path.dirname(os.path.realpath(__file__)))
        / "resources"
        / "test_configs"
        / "jobs_list2.csv"
    )

    PATH_TO_EXAMPLE_CSV_FILE3 = (
        Path(os.path.dirname(os.path.realpath(__file__)))
        / "resources"
        / "test_configs"
        / "jobs_list3.csv"
    )

    @patch("boto3.client")
    @patch("aind_data_transfer.jobs.basic_job.BasicJob.run_job")
    def test_load_configs(
        self, mock_run_job: MagicMock, mock_client: MagicMock
    ) -> None:
        """Tests configs are loaded correctly."""
        mock_resp1 = {
            "Parameter": {"Value": self.EXAMPLE_PARAM_STORE_RESPONSE}
        }
        mock_resp2 = {
            "Parameter": {"Value": json.dumps({"password": "some_password"})}
        }
        mock_resp3 = {
            "Parameter": {
                "Value": json.dumps(
                    {"CODEOCEAN_READWRITE_TOKEN": "some_token"}
                )
            }
        }

        mock_client.return_value.get_parameter.side_effect = [
            mock_resp1,
            mock_resp2,
            mock_resp3,
            mock_resp1,
            mock_resp2,
            mock_resp3,
        ]

        args = ["-j", str(self.PATH_TO_EXAMPLE_CSV_FILE)]
        dry_run_args = [
            "-j",
            str(self.PATH_TO_EXAMPLE_CSV_FILE),
            "--dry-run",
            "--compress-raw-data",
        ]
        jobs = GenericS3UploadJobList(args=args)
        dry_run_jobs = GenericS3UploadJobList(args=dry_run_args)

        self.assertFalse(jobs.job_list[0].job_configs.dry_run)
        self.assertTrue(dry_run_jobs.job_list[0].job_configs.dry_run)
        jobs.run_job()
        dry_run_jobs.run_job()

        expected_modalities_job0 = [
            ModalityConfigs(
                modality=Modality.ECEPHYS,
                source=Path(
                    "tests/resources/v0.6.x_neuropixels_multiexp_multistream"
                ),
            )
        ]

        self.assertEqual(
            expected_modalities_job0, jobs.job_list[0].job_configs.modalities
        )

        mock_client.assert_has_calls(
            [
                call("ssm"),
                call().get_parameter(
                    Name="/some/aws/param/store/name", WithDecryption=False
                ),
                call().close(),
                call("ssm"),
                call().get_parameter(
                    Name="/path/to/password", WithDecryption=True
                ),
                call().close(),
                call("ssm"),
                call().get_parameter(
                    Name="/path/to/token", WithDecryption=True
                ),
                call().close(),
                call("ssm"),
                call().get_parameter(
                    Name="/some/aws/param/store/name", WithDecryption=False
                ),
                call().close(),
                call("ssm"),
                call().get_parameter(
                    Name="/path/to/password", WithDecryption=True
                ),
                call().close(),
                call("ssm"),
                call().get_parameter(
                    Name="/path/to/token", WithDecryption=True
                ),
                call().close(),
            ]
        )
        mock_run_job.assert_has_calls([call(), call(), call(), call()])
        self.assertEqual(2, len(jobs.job_list))
        self.assertEqual(2, len(dry_run_jobs.job_list))

    @patch("boto3.client")
    @patch("aind_data_transfer.jobs.basic_job.BasicJob.run_job")
    def test_load_configs2(
        self, mock_run_job: MagicMock, mock_client: MagicMock
    ) -> None:
        """Tests configs are loaded correctly."""
        mock_resp1 = {
            "Parameter": {"Value": self.EXAMPLE_PARAM_STORE_RESPONSE}
        }
        mock_resp2 = {
            "Parameter": {"Value": json.dumps({"password": "some_password"})}
        }
        mock_resp3 = {
            "Parameter": {
                "Value": json.dumps(
                    {"CODEOCEAN_READWRITE_TOKEN": "some_token"}
                )
            }
        }

        mock_client.return_value.get_parameter.side_effect = [
            mock_resp1,
            mock_resp2,
            mock_resp3,
        ]

        args = ["-j", str(self.PATH_TO_EXAMPLE_CSV_FILE2)]
        jobs = GenericS3UploadJobList(args=args)

        # Default compress_raw_data should be false for non-ephys jobs
        self.assertFalse(jobs.job_list[0].job_configs.dry_run)
        # Default compress_raw_data should be true for ephys data
        self.assertFalse(jobs.job_list[1].job_configs.dry_run)
        self.assertTrue(jobs.job_list[2].job_configs.dry_run)
        self.assertFalse(jobs.job_list[3].job_configs.dry_run)
        jobs.run_job()
        # There are three jobs defined in the csv file
        mock_run_job.assert_has_calls([call(), call(), call(), call()])

    @patch("boto3.client")
    def test_load_configs3(self, mock_client: MagicMock) -> None:
        """Tests configs are loaded correctly."""
        mock_resp1 = {
            "Parameter": {"Value": self.EXAMPLE_PARAM_STORE_RESPONSE}
        }
        mock_resp2 = {
            "Parameter": {"Value": json.dumps({"password": "some_password"})}
        }
        mock_resp3 = {
            "Parameter": {
                "Value": json.dumps(
                    {"CODEOCEAN_READWRITE_TOKEN": "some_token"}
                )
            }
        }

        mock_client.return_value.get_parameter.side_effect = [
            mock_resp1,
            mock_resp2,
            mock_resp3,
            mock_resp1,
            mock_resp2,
            mock_resp3,
        ]

        args = ["-j", str(self.PATH_TO_EXAMPLE_CSV_FILE3)]
        jobs = GenericS3UploadJobList(args=args)
        job0 = jobs.job_list[0]
        job1 = jobs.job_list[1]
        self.assertEqual(
            ExperimentType.ECEPHYS, job0.job_configs.experiment_type
        )
        self.assertEqual(
            ExperimentType.CONFOCAL, job1.job_configs.experiment_type
        )
        expected_source = Path(
            "tests/resources/v0.6.x_neuropixels_multiexp_multistream"
        )
        self.assertTrue(
            (
                ModalityConfigs(
                    modality=Modality.ECEPHYS, source=expected_source
                )
                in job0.job_configs.modalities
            )
            and (
                ModalityConfigs(
                    modality=Modality.CONFOCAL, source=expected_source
                )
                in job0.job_configs.modalities
            )
        )
        self.assertTrue(
            (ExperimentType.ECEPHYS, job0.job_configs.experiment_type)
            and (ExperimentType.CONFOCAL, job1.job_configs.experiment_type)
        )
        self.assertIsNone(job0.job_configs.codeocean_process_capsule_id)
        self.assertIsNone(job1.job_configs.codeocean_process_capsule_id)

    @patch("boto3.client")
    def test_load_configs4(self, mock_client: MagicMock) -> None:
        """Tests configs with custom processing capsule id are loaded
        correctly."""
        mock_resp1 = {
            "Parameter": {"Value": self.EXAMPLE_PARAM_STORE_RESPONSE}
        }
        mock_resp2 = {
            "Parameter": {"Value": json.dumps({"password": "some_password"})}
        }
        mock_resp3 = {
            "Parameter": {
                "Value": json.dumps(
                    {"CODEOCEAN_READWRITE_TOKEN": "some_token"}
                )
            }
        }

        mock_client.return_value.get_parameter.side_effect = [
            mock_resp1,
            mock_resp2,
            mock_resp3,
            mock_resp1,
            mock_resp2,
            mock_resp3,
        ]

        args = ["-j", str(self.PATH_TO_EXAMPLE_CSV_FILE3)]
        jobs = GenericS3UploadJobList(args=args)
        job0 = jobs.job_list[0]
        job1 = jobs.job_list[1]
        self.assertEqual(
            ExperimentType.ECEPHYS, job0.job_configs.experiment_type
        )
        self.assertEqual(
            ExperimentType.CONFOCAL, job1.job_configs.experiment_type
        )
        expected_source = Path(
            "tests/resources/v0.6.x_neuropixels_multiexp_multistream"
        )
        self.assertTrue(
            (
                ModalityConfigs(
                    modality=Modality.ECEPHYS, source=expected_source
                )
                in job0.job_configs.modalities
            )
            and (
                ModalityConfigs(
                    modality=Modality.CONFOCAL, source=expected_source
                )
                in job0.job_configs.modalities
            )
        )
        self.assertTrue(
            (ExperimentType.ECEPHYS, job0.job_configs.experiment_type)
            and (ExperimentType.CONFOCAL, job1.job_configs.experiment_type)
        )
        self.assertTrue(
            ("xyz-123", job0.job_configs.codeocean_process_capsule_id)
            and ("zyx-456", job1.job_configs.codeocean_process_capsule_id)
        )

    @patch("boto3.client")
    @patch("aind_data_transfer.jobs.basic_job.BasicJob.run_job")
    @patch("logging.error")
    def test_errors_skipped(
        self,
        mock_log_error: MagicMock,
        mock_run_job: MagicMock,
        mock_client: MagicMock,
    ) -> None:
        """Tests that jobs with errors are skipped instead of terminating
        list."""
        mock_resp1 = {
            "Parameter": {"Value": self.EXAMPLE_PARAM_STORE_RESPONSE}
        }
        mock_resp2 = {
            "Parameter": {"Value": json.dumps({"password": "some_password"})}
        }
        mock_resp3 = {
            "Parameter": {
                "Value": json.dumps(
                    {"CODEOCEAN_READWRITE_TOKEN": "some_token"}
                )
            }
        }

        mock_client.return_value.get_parameter.side_effect = [
            mock_resp1,
            mock_resp2,
            mock_resp3,
        ]

        args = ["-j", str(self.PATH_TO_EXAMPLE_CSV_FILE2)]
        mock_run_job.side_effect = [
            None,
            Exception("Test"),
            None,
            Exception("Test"),
        ]
        jobs = GenericS3UploadJobList(args=args)

        jobs.run_job()
        mock_log_error.assert_has_calls(
            [
                call(
                    "There was a problem processing job 2 of 4. "
                    "Skipping for now. Error: Test"
                ),
                call(
                    "There was a problem processing job 4 of 4. "
                    "Skipping for now. Error: Test"
                ),
                call("There were errors processing the following jobs: "),
                call(
                    "There was a problem processing job 2 of 4. " "Error: Test"
                ),
                call(
                    "There was a problem processing job 4 of 4. " "Error: Test"
                ),
            ]
        )


if __name__ == "__main__":
    unittest.main()



from enum import Enum

from numcodecs import Blosc


class ImagingCompressors:
    class Compressors(Enum):
        """Enum for compression algorithms a user can select"""

        blosc = Blosc.codec_id

    compressors = [member.value for member in Compressors]

    @staticmethod
    def get_compressor(compressor_name, **kwargs):
        """
        Retrieve a compressor for a given name and optional kwargs.
        Args:
            compressor_name (str): Matches one of the names Compressors enum
            **kwargs (dict): Options to pass into the Compressor
        Returns:
            An instantiated compressor class.
        """
        if compressor_name == ImagingCompressors.Compressors.blosc.name:
            return Blosc(**kwargs)
        else:
            raise Exception(
                f"Unknown compressor. Please select one of "
                f"{ImagingCompressors.compressors}"
            )



"""Loads Ephys job configurations"""
import argparse
import json
import logging
import os
import re
from pathlib import Path
from warnings import warn

import yaml
from botocore.exceptions import ClientError
from numcodecs import Blosc

from aind_data_transfer.readers.ephys_readers import EphysReaders
from aind_data_transfer.util.s3_utils import get_secret

warn(
    f"The module {__name__} is deprecated and will be removed in future "
    f"versions.",
    DeprecationWarning,
    stacklevel=2,
)


class EphysJobConfigurationLoader:
    """Class to handle loading ephys job configs"""

    DEFAULT_AWS_REGION = "us-west-2"
    # Not the actual passwords, just the key name to retrieve it from aws
    # secrets manager
    SERVICE_ENDPOINT_KEY = "service_endpoints"
    VIDEO_ENCRYPTION_KEY_NAME = "video_encryption_password"
    CODE_OCEAN_API_TOKEN_NAME = "codeocean-api-token"
    # TODO: Is there a better way to do this?
    CODE_REPO_LOCATION = (
        "https://github.com/AllenNeuralDynamics/aind-data-transfer"
    )

    @staticmethod
    def _get_endpoints(s3_region: str) -> dict:
        """
        If the service endpoints aren't set in the sys args, then this method
        will try to pull them from aws secrets manager. It's static since it's
        being called before the job is created.
        Parameters
        ----------
        s3_region : str

        Returns
        -------
        dict
          Will return an empty dictionary if the service endpoints are not set
          in sys args, or if they can't be pulled from aws secrets manager.

        """
        try:
            s3_secret_name = EphysJobConfigurationLoader.SERVICE_ENDPOINT_KEY
            get_secret(s3_secret_name, s3_region)
            endpoints = json.loads(get_secret(s3_secret_name, s3_region))
        except ClientError as e:
            logging.warning(
                f"Unable to retrieve aws secret: "
                f"{EphysJobConfigurationLoader.SERVICE_ENDPOINT_KEY}"
            )
            logging.debug(e.response)
            endpoints = {}
        return endpoints

    def __remove_none(self, data):
        """Remove keys whose value is None."""
        if isinstance(data, dict):
            return {
                k: self.__remove_none(v)
                for k, v in data.items()
                if v is not None
            }
        else:
            return data

    @staticmethod
    def __parse_compressor_configs(configs):
        """Util method to map a string to class attribute"""
        try:
            compressor_name = configs["compress_data_job"]["compressor"][
                "compressor_name"
            ]
        except KeyError:
            compressor_name = None
        try:
            compressor_kwargs = configs["compress_data_job"]["compressor"][
                "kwargs"
            ]
        except KeyError:
            compressor_kwargs = None
        if (
            compressor_name
            and compressor_name == Blosc.codec_id
            and compressor_kwargs
            and "shuffle" in compressor_kwargs
        ):
            shuffle_str = configs["compress_data_job"]["compressor"]["kwargs"][
                "shuffle"
            ]
            shuffle_val = getattr(Blosc, shuffle_str)
            configs["compress_data_job"]["compressor"]["kwargs"][
                "shuffle"
            ] = shuffle_val

    @staticmethod
    def __resolve_endpoints(configs):  # noqa: C901
        """
        Only the raw data source needs to be provided as long as the base dir
        name is formatted correctly. If the dest_data_dir and cloud endpoints
        are not set in the conf file, they will be created automatically based
        on the name of the raw_data_source.
        Args:
            configs (dic): Configurations

        Returns:
            None, modifies the base configs in place
        """
        raw_data_folder = Path(configs["endpoints"]["raw_data_dir"]).name

        dest_data_dir = configs["endpoints"]["dest_data_dir"]
        reg_match_subject_datetime = re.match(
            EphysReaders.SourceRegexPatterns.subject_datetime.value,
            raw_data_folder,
        )
        reg_match_ecephys_subject_datetime = re.match(
            EphysReaders.SourceRegexPatterns.ecephys_subject_datetime.value,
            raw_data_folder,
        )
        if dest_data_dir is None and reg_match_subject_datetime is not None:
            configs["endpoints"]["dest_data_dir"] = (
                "ecephys_" + raw_data_folder
            )
        if (
            configs["data"].get("subject_id") is None
            and reg_match_subject_datetime is not None
        ):
            configs["data"]["subject_id"] = reg_match_subject_datetime.group(1)
        if (
            dest_data_dir is None
            and reg_match_ecephys_subject_datetime is not None
        ):
            configs["endpoints"]["dest_data_dir"] = raw_data_folder
        if (
            configs["data"].get("subject_id") is None
            and reg_match_ecephys_subject_datetime is not None
        ):
            configs["data"][
                "subject_id"
            ] = reg_match_ecephys_subject_datetime.group(1)

        if configs["endpoints"]["s3_prefix"] is None:
            dest_data_folder = Path(configs["endpoints"]["dest_data_dir"]).name
            configs["endpoints"]["s3_prefix"] = dest_data_folder

        if configs["endpoints"]["gcp_prefix"] is None:
            dest_data_folder = Path(configs["endpoints"]["dest_data_dir"]).name
            configs["endpoints"]["gcp_prefix"] = dest_data_folder

        if configs["trigger_codeocean_job"]["job_type"] is None:
            configs["trigger_codeocean_job"]["job_type"] = configs["data"][
                "name"
            ]

        if configs["trigger_codeocean_job"]["bucket"] is None:
            configs["trigger_codeocean_job"]["bucket"] = configs["endpoints"][
                "s3_bucket"
            ]

        if configs["trigger_codeocean_job"]["prefix"] is None:
            configs["trigger_codeocean_job"]["prefix"] = configs["endpoints"][
                "s3_prefix"
            ]

        if configs["aws_secret_names"]["region"] is None:
            configs["aws_secret_names"][
                "region"
            ] = EphysJobConfigurationLoader.DEFAULT_AWS_REGION

        # Not the actual password, just the key used to retrieve it
        if configs["aws_secret_names"]["video_encryption_password"] is None:
            configs["aws_secret_names"][
                "video_encryption_password"
            ] = EphysJobConfigurationLoader.VIDEO_ENCRYPTION_KEY_NAME

        if configs["aws_secret_names"]["code_ocean_api_token_name"] is None:
            configs["aws_secret_names"][
                "code_ocean_api_token_name"
            ] = EphysJobConfigurationLoader.CODE_OCEAN_API_TOKEN_NAME

        if configs["endpoints"]["code_repo_location"] is None:
            configs["endpoints"][
                "code_repo_location"
            ] = EphysJobConfigurationLoader.CODE_REPO_LOCATION

        if (
            (configs["endpoints"]["codeocean_domain"] is None)
            or (configs["endpoints"]["metadata_service_url"] is None)
            or (configs["trigger_codeocean_job"]["capsule_id"] is None)
        ):
            service_endpoints = EphysJobConfigurationLoader._get_endpoints(
                configs["aws_secret_names"]["region"]
            )
            if configs["endpoints"]["codeocean_domain"] is None:
                configs["endpoints"]["codeocean_domain"] = service_endpoints[
                    "codeocean_domain"
                ]
            if configs["endpoints"]["metadata_service_url"] is None:
                configs["endpoints"][
                    "metadata_service_url"
                ] = service_endpoints["metadata_service_url"]

            if configs["trigger_codeocean_job"]["capsule_id"] is None:
                configs["trigger_codeocean_job"][
                    "capsule_id"
                ] = service_endpoints["codeocean_trigger_capsule"]

    @staticmethod
    def __resolve_logging(configs: dict) -> None:
        """
        Resolves logging config in place
        Parameters
        ----------
        configs : dict
          Configurations

        Returns
        -------
        None
        """

        if configs["logging"]["level"] is None:
            configs["logging"]["level"] = "INFO"
        if os.getenv("LOG_LEVEL"):
            configs["logging"]["level"] = os.getenv("LOG_LEVEL")

    def load_configs(self, sys_args):
        """Load yaml config at conf_src Path as python dict"""
        parser = argparse.ArgumentParser()
        parser.add_argument(
            "-c", "--conf-file-location", required=True, type=str
        )
        parser.add_argument(
            "-r", "--raw-data-source", required=False, type=str
        )
        parser.add_argument(
            "-b", "--behavior-directory", required=False, type=str
        )
        args = parser.parse_args(sys_args)
        conf_src = args.conf_file_location
        with open(conf_src) as f:
            raw_config = yaml.load(f, Loader=yaml.SafeLoader)
        if args.raw_data_source is not None:
            raw_config["endpoints"]["raw_data_dir"] = args.raw_data_source
        if args.behavior_directory is not None:
            raw_config["endpoints"][
                "behavior_directory"
            ] = args.behavior_directory
        self.__resolve_endpoints(raw_config)
        self.__resolve_logging(raw_config)
        config_without_nones = self.__remove_none(raw_config)
        self.__parse_compressor_configs(config_without_nones)
        return config_without_nones



import math
import os
import tempfile
import unittest
from pathlib import Path

import h5py
import numpy as np
import tifffile
import zarr
from distributed import Client
from parameterized import parameterized

from aind_data_transfer.transformations.ome_zarr import write_files, write_folder
from aind_data_transfer.util.io_utils import ImarisReader


def _write_test_tiffs(folder, n=4, shape=(64, 128, 128)):
    for i in range(n):
        a = np.ones(shape, dtype=np.uint16)
        tifffile.imwrite(os.path.join(folder, f"data_{i}.tif"), a, imagej=True)


def _write_test_h5(folder, n=4, shape=(64, 128, 128)):
    for i in range(n):
        a = np.ones(shape, dtype=np.uint16)
        with h5py.File(os.path.join(folder, f"data_{i}.h5"), "w") as f:
            f.create_dataset(ImarisReader.DEFAULT_DATA_PATH, data=a, chunks=True)
            # Write origin metadata
            dataset_info = f.create_group("DataSetInfo/Image")
            dataset_info.attrs["ExtMin0"] = np.array(
                ["1", "0", "0"], dtype="S"
            )
            dataset_info.attrs["ExtMin1"] = np.array(
                ["2", "0", "0"], dtype="S"
            )
            dataset_info.attrs["ExtMin2"] = np.array(
                ["3", "0", "0"], dtype="S"
            )
            dataset_info.attrs["X"] = np.array(
                list(str(shape[2])), dtype="S"
            )
            dataset_info.attrs["Y"] = np.array(
                list(str(shape[1])), dtype="S"
            )
            dataset_info.attrs["Z"] = np.array(
                list(str(shape[0])), dtype="S"
            )


class TestOmeZarr(unittest.TestCase):
    """Tests methods defined in aind_data_transfer.transcode.ome_zarr class"""

    def setUp(self):
        self._temp_dir = tempfile.TemporaryDirectory()
        self._image_dir = Path(self._temp_dir.name) / "images"
        os.makedirs(self._image_dir, exist_ok=True)
        self.tiff_dir = self._image_dir / "tiff"
        self.h5_dir = self._image_dir / "h5"
        os.mkdir(self.tiff_dir)
        os.mkdir(self.h5_dir)
        _write_test_tiffs(self.tiff_dir)
        _write_test_h5(self.h5_dir)
        self._client = Client()

    def tearDown(self):
        self._temp_dir.cleanup()
        self._client.close()

    def _check_multiscales_arrays(
        self, z, full_shape, actual_keys, n_levels, scale_factor
    ):
        # Test arrays across resolution levels
        for key in actual_keys:
            for lvl in range(n_levels):
                a = z[key + f"/{lvl}"]
                expected_shape_at_lvl = (
                    1,
                    1,
                    int(math.ceil(full_shape[2] / (scale_factor**lvl))),
                    int(math.ceil(full_shape[3] / (scale_factor**lvl))),
                    int(math.ceil(full_shape[4] / (scale_factor**lvl))),
                )
                self.assertEqual(expected_shape_at_lvl, a.shape)
                self.assertTrue(a.nbytes_stored > 0)

    def _check_zarr_attributes(
        self, z, voxel_size, scale_factor, has_translation
    ):
        for key in z.keys():
            tile_group = z[key]
            attrs = dict(tile_group.attrs)
            expected_omero_metadata = {
                "channels": [
                    {
                        "active": True,
                        "coefficient": 1,
                        "color": "000000",
                        "family": "linear",
                        "inverted": False,
                        "label": f"Channel:{key}:0",
                        "window": {
                            "end": 1.0,
                            "max": 1.0,
                            "min": 0.0,
                            "start": 0.0,
                        },
                    }
                ],
                "id": 1,
                "name": f"{key}",
                "rdefs": {"defaultT": 0, "defaultZ": 32, "model": "color"},
                "version": "0.4",  # TODO: test against version?
            }
            actual_omero_metadata = attrs["omero"]
            self.assertEqual(
                expected_omero_metadata["channels"],
                actual_omero_metadata["channels"],
            )
            self.assertEqual(
                expected_omero_metadata["id"], actual_omero_metadata["id"]
            )
            self.assertEqual(
                expected_omero_metadata["name"], actual_omero_metadata["name"]
            )
            self.assertEqual(
                expected_omero_metadata["rdefs"],
                actual_omero_metadata["rdefs"],
            )

            expected_axes_metadata = [
                {"name": "t", "type": "time", "unit": "millisecond"},
                {"name": "c", "type": "channel"},
                {"name": "z", "type": "space", "unit": "micrometer"},
                {"name": "y", "type": "space", "unit": "micrometer"},
                {"name": "x", "type": "space", "unit": "micrometer"},
            ]
            self.assertEqual(
                expected_axes_metadata, attrs["multiscales"][0]["axes"]
            )

            datasets_metadata = attrs["multiscales"][0]["datasets"]
            for i, ds_metadata in enumerate(datasets_metadata):
                expected_transform = {
                    "coordinateTransformations": [
                        {
                            "scale": [
                                1.0,
                                1.0,
                                voxel_size[0] * (scale_factor**i),
                                voxel_size[1] * (scale_factor**i),
                                voxel_size[2] * (scale_factor**i),
                            ],
                            "type": "scale",
                        }
                    ],
                    "path": f"{i}",
                }
                if has_translation:
                    expected_transform["coordinateTransformations"].append(
                        {
                            "translation": [0, 0, 300.0, 200.0, 100.0],
                            "type": "translation",
                        }
                    )
                self.assertEqual(expected_transform, datasets_metadata[i])

            self.assertEqual(f"/{key}", attrs["multiscales"][0]["name"])

    @parameterized.expand(["h5", "tiff"])
    def test_write_files(self, file_type):
        image_dir = None
        if file_type == "h5":
            image_dir = self.h5_dir
        elif file_type == "tiff":
            image_dir = self.tiff_dir

        files = list(sorted([image_dir / f for f in image_dir.iterdir()]))

        shape = (64, 128, 128)
        out_zarr = os.path.join(self._temp_dir.name, "ome.zarr")
        n_levels = 4
        scale_factor = 2.0
        voxel_size = [1.0, 1.0, 1.0]
        metrics = write_files(
            files, out_zarr, n_levels, scale_factor, voxel_size=voxel_size
        )

        self.assertEqual(len(metrics), len(files))

        z = zarr.open(out_zarr, "r")

        # Test group for each written image
        expected_keys = {"data_0.zarr", "data_1.zarr", "data_2.zarr", "data_3.zarr"}
        actual_keys = set(z.keys())
        self.assertEqual(expected_keys, actual_keys)

        expected_shape = (1, 1, *shape)
        self._check_multiscales_arrays(
            z, expected_shape, actual_keys, n_levels, scale_factor
        )

        self._check_zarr_attributes(
            z, voxel_size, scale_factor, has_translation=(file_type == "h5")
        )

    @parameterized.expand(["h5", "tiff"])
    def test_write_folder(self, file_type):
        image_dir = None
        if file_type == "h5":
            image_dir = self.h5_dir
        elif file_type == "tiff":
            image_dir = self.tiff_dir

        shape = (64, 128, 128)
        out_zarr = os.path.join(self._temp_dir.name, "ome.zarr")
        n_levels = 4
        scale_factor = 2.0
        voxel_size = [1.0, 1.0, 1.0]

        # create a dummy file to exclude from conversion
        tifffile.imwrite(image_dir / "dummy.tif", np.zeros(shape))
        exclude = ["*dummy*"]

        write_folder(
            image_dir,
            out_zarr,
            n_levels,
            scale_factor,
            voxel_size=voxel_size,
            exclude=exclude,
        )

        z = zarr.open(out_zarr, "r")

        # Test that all images were written except exclusion
        expected_keys = {"data_0.zarr", "data_1.zarr", "data_2.zarr", "data_3.zarr"}
        actual_keys = set(z.keys())
        self.assertEqual(expected_keys, actual_keys)

        expected_shape = (1, 1, *shape)
        self._check_multiscales_arrays(
            z, expected_shape, actual_keys, n_levels, scale_factor
        )

        self._check_zarr_attributes(
            z, voxel_size, scale_factor, has_translation=(file_type == "h5")
        )


if __name__ == "__main__":
    unittest.main()



import logging
from typing import List, Union

import google.auth
import google.auth.exceptions
from google.cloud import storage
from google.cloud.storage import Client

from aind_data_transfer.util.file_utils import (
    collect_filepaths,
    make_cloud_paths,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def create_client() -> Client:
    try:
        logger.info("Retrieving GCS credentials")
        credentials, project = google.auth.default()
        logger.info("Creating client")
        client = storage.Client(
            credentials=credentials,
            project=project,
        )
    except Exception as e:
        logger.error(
            f"Error while authenticating Google client. Verify json file. \n{e}"
        )
        raise
    return client


class GCSUploader:
    def __init__(self, bucket_name: str):
        self.client = create_client()
        try:
            self.bucket = self.client.get_bucket(bucket_name)
        except google.cloud.exceptions.NotFound:
            logger.error(f"Bucket {bucket_name} not found")
            raise

    def upload_file(
        self,
        filepath: str,
        gcs_key: str,
        timeout: float = None,
        chunk_size: int = 64 * 1024 * 1024,
    ) -> bool:
        """Upload a single file to gcs.
        Args:
            filepath (str): absolute path to file to upload.
            gcs_bucket (str): name of gcs bucket
            gcs_key (str): location relative to bucket to store blob
            timeout (float): upload timeout
            chunk_size (int): upload file in chunks of this size
        Returns:
            A list of filepaths for failed uploads
        """
        blob = self.bucket.blob(gcs_key)
        blob.chunk_size = chunk_size
        try:
            blob.upload_from_filename(filepath, timeout=timeout)
            return True
        except Exception as e:
            logger.error(f"Error uploading file {filepath}\n{e}")
            return False

    def upload_files(
        self,
        filepaths: List[str],
        gcs_path: Union[str, List[str]],
        root: str = None,
        chunk_size: int = 64 * 1024 * 1024,
    ) -> List[str]:
        """Upload a list of files to gcs.
        Args:
            filepaths (list): absolute paths of files to upload.
            gcs_bucket (str): name of gcs bucket
            gcs_path (str or list): if a string, location relative to bucket to store objects.
                                    If a list, the cloud path for each file in filepaths.
            root (str): root directory shared by all files in filepaths.
                        If None, all files will be stored as a flat list
                        under gcs_folder. Default is None.
            chunk_size (int): upload each file in chunks of this size
        Returns:
            A list of filepaths for failed uploads
        """
        if isinstance(gcs_path, str):
            gcs_paths = make_cloud_paths(filepaths, gcs_path, root=root)
        else:
            try:
                _ = iter(gcs_path)
                gcs_paths = gcs_path
            except TypeError:
                logger.error(
                    f"Expected either a str or iterable, got {type(gcs_path)}"
                )
                raise
        failed_uploads = []
        for fpath, gpath in zip(filepaths, gcs_paths):
            if not self.upload_file(fpath, gpath, chunk_size=chunk_size):
                failed_uploads.append(fpath)
        return failed_uploads

    def upload_folder(
        self,
        folder: str,
        gcs_folder: str,
        recursive: bool = True,
        chunk_size: int = 64 * 1024 * 1024,
    ) -> List[str]:
        """Upload a directory to gcs.
        Args:
            folder (str): absolute path of the folder to upload.
            gcs_folder (str): location relative to bucket to store objects
            recursive (boolean): upload all sub-directories
            chunk_size (int): upload each file in chunks of this size
        Returns:
            A list of filepaths for failed uploads
        """
        return self.upload_files(
            collect_filepaths(folder, recursive),
            gcs_folder,
            root=folder,
            chunk_size=chunk_size,
        )



"""Unit tests for ephys readers module."""
import os
import unittest
from pathlib import Path

from aind_data_transfer.readers.ephys_readers import EphysReaders, DataReader

TEST_DIR = Path(os.path.dirname(os.path.realpath(__file__)))
RESOURCES_DIR = TEST_DIR / "resources"


class TestEphysReaders(unittest.TestCase):

    open_ephys_dir = RESOURCES_DIR / "v0.6.x_neuropixels_multiexp_multistream"

    nidaq_rec_prefix = (
        "'recording': OpenEphysBinaryRecordingExtractor: "
        "8 channels - 30.0kHz - 1 segments - 100 samples \n"
        "                                   0.00s (3.33 ms) - int16 dtype - 1.56 KiB, "
        "'experiment_name': 'experiment"
    )
    nidaq_rec_suffix = (
        "'stream_name': 'Record Node 101#NI-DAQmx-103.PXIe-6341'"
    )

    neuropix_rec_prefix = (
        "'recording': OpenEphysBinaryRecordingExtractor: "
        "384 channels - 30.0kHz - 1 segments - 100 samples \n"
        "                                   0.00s (3.33 ms) - int16 dtype - 75.00 KiB, "
        "'experiment_name': 'experiment"
    )
    neuropix_rec_suffix = (
        "'stream_name': 'Record Node 101#Neuropix-PXI-100.Probe"
    )

    expected_read_blocks_info = [
        f"{{{nidaq_rec_prefix}1', {nidaq_rec_suffix}}}",
        f"{{{neuropix_rec_prefix}1', {neuropix_rec_suffix}B'}}",
        f"{{{neuropix_rec_prefix}1', {neuropix_rec_suffix}C'}}",
        f"{{{nidaq_rec_prefix}3', {nidaq_rec_suffix}}}",
        f"{{{neuropix_rec_prefix}3', {neuropix_rec_suffix}B'}}",
        f"{{{neuropix_rec_prefix}3', {neuropix_rec_suffix}C'}}",
        f"{{{nidaq_rec_prefix}6', {nidaq_rec_suffix}}}",
        f"{{{neuropix_rec_prefix}6', {neuropix_rec_suffix}B'}}",
        f"{{{neuropix_rec_prefix}6', {neuropix_rec_suffix}C'}}",
    ]

    def test_read(self):
        openephys_reader = DataReader.OPENEPHYS.value
        read_blocks = EphysReaders.get_read_blocks(
            openephys_reader, self.open_ephys_dir
        )
        read_blocks_info = [str(r) for r in read_blocks]
        self.assertEqual(self.expected_read_blocks_info, read_blocks_info)

    def test_read_fails(self):
        with self.assertRaises(Exception):
            list(
                EphysReaders.get_read_blocks(
                    "Made up readers", self.open_ephys_dir
                )
            )


if __name__ == "__main__":
    unittest.main()



import logging
import os
import socket
from typing import Optional, Tuple

import distributed
from dask_mpi import initialize
from distributed import Client, LocalCluster

logging.basicConfig(format="%(asctime)s %(message)s", datefmt="%Y-%m-%d %H:%M")
LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


def log_dashboard_address(
        client: distributed.Client,
        login_node_address: str = "hpc-login"
) -> None:
    """
    Logs the terminal command required to access the Dask dashboard

    Args:
        client: the Client instance
        login_node_address: the address of the cluster login node
    """
    host = client.run_on_scheduler(socket.gethostname)
    port = client.scheduler_info()['services']['dashboard']
    user = os.getenv("USER")
    LOGGER.info(
        f"To access the dashboard, run the following in a terminal: ssh -L {port}:{host}:{port} {user}@"
        f"{login_node_address} "
    )


def get_client(
        deployment: str = "local",
        worker_options: Optional[dict] = None,
        n_workers: int = 1
) -> Tuple[distributed.Client, int]:
    """
    Create a distributed Client

    Args:
        deployment: the type of deployment. Either "local" or "slurm"
        worker_options: a dictionary of options to pass to the worker class
        n_workers: the number of workers (only applies to "local" deployment)

    Returns:
        the distributed Client and number of workers
    """
    if deployment == "slurm":
        if worker_options is None:
            worker_options = {}
        slurm_job_id = os.getenv("SLURM_JOBID")
        if slurm_job_id is None:
            raise Exception(
                "SLURM_JOBID environment variable is not set. Are you running under SLURM?"
            )
        initialize(
            nthreads=int(os.getenv("SLURM_CPUS_PER_TASK", 1)),
            local_directory=f"/scratch/fast/{slurm_job_id}",
            worker_class="distributed.nanny.Nanny",
            worker_options=worker_options
        )
        client = Client()
        log_dashboard_address(client)
        n_workers = int(os.getenv("SLURM_NTASKS"))
    elif deployment == "local":
        import platform
        use_procs = False if platform.system() == "Windows" else True
        cluster = LocalCluster(n_workers=n_workers, processes=use_procs, threads_per_worker=1)
        client = Client(cluster)
    else:
        raise NotImplementedError
    return client, n_workers
